//
// Copyright 1996, David E. Bodenstab
// All rights reserved.
//
// Generate the data structures used by emit() for XCOM.  These include
// the TEMPLATE array, the i486Opcodes array, and the various LITERAL
// declarations.  Note that during the Intel porting effort, several
// variations of the data structures existed, finally evolving to the
// present form.  This program generates the data structures for the
// current algorithm used by the emit() routine.  The intermediate data
// structures were at one point machine generated, but the program no
// longer exists.  However, there should be no reason to want to modify
// one of the intermediate revisions of XCOM.XPL, since the latest
// revision would be the logical point to begin further modifications.
//
// The instruction table immediately follows this source, and the program
// is designed to read the source file, skipping forward until the
// instruction table is reached.
//
// Run this program either of two ways.  The easiest:
//
//      gmake template.data
//
// Without the use of gmake:
//
//      xpl template.xpl
//      xplsm -p template --input template.xpl --output3 template.data
//
//
// The instruction table consists of instruction templates, one per line.
// A comment is anything following a semicolon.  Blank lines are ignored.
// Four fields (three mandatory) separated by blanks may be provided.
//
// The first field is the instruction template.  It consists of the 
// instruction name and up to three operand types separated by commas.
// An alternate LITERAL name for the instruction may be provided by
// following the name by =alternate; this is used for `call' which
// is a reserved word in XPL.
//
// The second field is the opcode used when the instruction is printed
// by using the $E compiler toggle.  The opcode is a string enclosed
// within single quotes (an XPL character string.)
//
// The third field is the hexadecimal opcode for the instruction.  If
// the operand type is REL, then two hexadecimal opcodes are provided
// separated by a comma.  Each opcode is enclosed in double quotes (an
// XPL bit constant.)
//
// The fourth (optional) field are any flag values used by the emit()
// routine.  Multiple values are separated by commas.
//
// Consult XCOM.XPL, the emit() function, for the definitions of the
// various operand and flag values.
//
//
// Most instructions are not generated by XCOM itself, but the template
// table is reasonably complete to allow INLINE() to be used to generate
// these instructions.
//

declare lit  literally 'literally',
        char lit       'character',
        proc lit       'procedure',
        fix  lit       'fixed',
        init lit       'initial',
        dcl  lit       'declare',

        forever lit 'while "1"',

        Stdin     lit 'input(0)',
        Stdout    lit 'output(0)',
        Stderr    lit 'output(2)';


dcl bool  lit 'bit(8)',
    true  lit '1',
    false lit '0';


//
// The TEMPLATE() BIT(56) array is encoded as:
//
//     XXXXXXXX XXXXXXXX XXXXXXXX ........ ........ ........ ........  code word
//     XXXXXXX. ........ ........                                        instruction
//     .......X XXXXXXXX XXXXXXXX                                        op1,op2,op3
//     ........ ........ ........ XXXXXXXX XXXXXXXX ........ ........  opcode or relops-index
//     ........ ........ ........ ........ ........ XXXXXXXX ........  flags
//     ........ ........ ........ ........ ........ ........ XXXXXXXX  opcode index
//
// Code word is:
//
//     +---------------+-----------------+---------------+-----+
//     |    opcode     |       op1       |       op2     | op3 |
//     | _ _ _ _ _ _ _ | _ _ _ _ _ x x x | _ _ _ _ x x x | _ _ |
//     |       7       |     5       3   |    4      3   |  2  | == 24
//     +---------------+-----------------+---------------+-----+
//

dcl MaxTEMPLATE lit '1000';
dcl TEMPLATE(MaxTEMPLATE) bit(56);
dcl nTemplate fix init(0);

dcl INST(MaxTEMPLATE) char;
dcl nInst fix init(0);

dcl MaxRELOPCODES lit '10';
dcl RELOPCODES(MaxRELOPCODES) bit(16);
dcl nRelopcodes fix init(0);

dcl MaxOPCODES lit '2000';
dcl OPCODES(MaxOPCODES) char;
dcl nOpcodes fix init(0);


dcl temp(2000) bit(16);     // Must match maximum size of TEMPLATE, INST and OPCODES
dcl newx(2000) bit(16);     // Must match maximum size of TEMPLATE, INST and OPCODES


//
// XPL version of strspn(3C)
//
// The strspn() function spans the initial part of the string STRING
// as long as the characters from STRING occur in string CHARSET.
// The strspn() function returns the number of characters spanned.
//
strspn: proc( string, charset ) fix;
  dcl (string, charset) char;
  dcl (n, p, pC, c) fix;

  n = 0;
  do p = 0 to length(string) - 1;
    pC = byte(string,p);

    do c = 0 to length(charset) - 1;
      if pC = byte(charset,c) then
        goto gotone;
    end;
    return n;

  gotone:
    n = n + 1;
  end;

  return n;
end strspn;


//
// XPL version of strcspn(3C)
//
// The strcspn() function spans the initial part of the string STRING
// as long as the characters from STRING do not occur in string CHARSET.
// The strcspn() function returns the number of characters spanned.
//
strcspn: proc( string, charset ) fix;
  dcl (string, charset) char;
  dcl (n, p, pC, c) fix;

  n = 0;
  do p = 0 to length(string) - 1;
    pC = byte(string,p);

    do c = 0 to length(charset) - 1;
      if pC = byte(charset,c) then
        return n;
    end;

    n = n + 1;
  end;

  return n;
end strcspn;


//
// XPL version of strtok(3C)
//

dcl NULL char init('NULL');

strtok: proc( string, delim ) char;
  dcl (string, delim, r) char;
  dcl n fix;

  /* Compare descriptors themselves to detect the `NULL' pointer */
  if inline(mov,r32,m32,none, 0, string) = inline(mov,r32,m32,none, 1, NULL) then
    if length(r) = 0 then
      return '';
    else
      string = r;

  n = strspn( string, delim );
  string = substr( string, n );

  n = strcspn( string, delim );
  if n = 0 then
    do;
      r = '';
      return r;
    end;
    
  r = substr( string, n );

  return substr( string, 0, n );
end strtok;


//
// XPL version of strchr(3C)
//
strchr: proc( string, c ) fix;
  dcl string char;
  dcl (c, i) fix;

  do i = 0 to length(string) - 1;
    if byte(string,i) = c then
      return i;
  end;

  return -1;
end strchr;


//
// XPL version of strcmp(3C)
//
strcmp: proc( a, b ) bit(1);
  dcl (a, b) char, (la, lb) fix;

  la = length(a);
  lb = length(b);

  if la = lb then
    do; /* XPL is honest in this case */
      if a = b then
        return 0;
      else
        if a < b then
          return -1;
        else
          return 1;
    end;

  if la > lb then
    do;
      if substr(a,0,lb) >= b then
        return 1;
      else
        return -1;
    end;
  else
    do;
      if a <= substr(b,0,la) then
        return -1;
      else
        return 1;
    end;
end strcmp;


//
// Convert STRING to decimal
//
atoi: proc( string ) fix;
  dcl string char;
  dcl (i, c, v) fix;

  v = 0;
  do i = 0 to length(string) - 1;
    c = byte( string, i );

    if byte('0') <= c & c <= byte('9') then
      v = v * 10 + (c - byte('0'));
    else
      return v;
  end;

  return v;
end atoi;


//
// Convert STRING to hexidecimal
//
atox: proc( string ) fix;
  dcl string char;
  dcl (i, c, v) fix;

  v = 0;
  do i = 0 to length(string) - 1;
    c = byte( string, i );

    if byte('0') <= c & c <= byte('9') then
      v = shl( v, 4 ) + (c - byte('0'));
    else
      if byte('a') <= c & c <= byte('f') then
        v = shl( v, 4 ) + (c - (byte('a') - 10));
      else
        if byte('A') <= c & c <= byte('F') then
          v = shl( v, 4 ) + (c - (byte('A') - 10));
        else
          return v;
  end;

  return v;
end atox;


// Translate a string to upper case
//
toUpper: proc( s ) char;
  dcl (s, t) char;
  dcl (c, i) fix;

  t = substr( ' ' || s, 1 );  // Make a copy of the string

  do i = 0 to length(t);
    c = byte(t,i);

    if byte('a') <= c & c <= byte('z') then
      byte(t,i) = c - (byte('a') - byte('A'));
  end;

  return t;
end toUpper;


//
// Convert a NUMBER to hex, returning it as a string of WIDTH characters
//
Xformat: proc(number, width) char;
  /* Warning, Xformat() returns a descriptor pointing into T; the value */
  /* of T will change upon the next call to Xformat() */
  dcl (number,width,i,k) fixed, t char init('        ');

  do k = 0 to 7;
    i = byte( '0123456789ABCDEF', number & "0F" );
    byte( t, 7-k ) = i;
    number = shr( number, 4 );
  end;

  if width <= 8 then return substr(t,8-width);

  return substr('0000000000000000000000000', 0, width-8) || t;
end Xformat;


//
// Pad a STRING with blanks to WIDTH characters
//
Pad: proc( string, width ) char;
  dcl string char, (width, l) fix;
  dcl x50 char init( '                                                  ' );

  l = length( string );

  if l >= width then return string;

  do while width - l > 50;
    string = string || x50;
    l = l + 50;
  end;

  return string || substr( x50, 0, width-l );
end Pad;


//
// Encode an instruction
//
Inst: proc( op, a1, a2, a3 ) fix;
  dcl (op, a1, a2, a3) fix;
  return shl(op,17) | shl(a1,12) | shl(a2,5) | (a3 & "03");
end Inst;


//
// Match operand and flag keywords
//
dcl Op1(17) char init('eax,al',''  ,'',''   ,'ah','none','imm8','imm32',
                      'r8','r16','r32','imm16','m8','m16','m32','rel8','rel32','rel');
dcl Op2(15) char init('eax',   'cl','','ebx',''  ,'none','imm8','imm32',
                      'r8','r16','r32','imm16','m8','m16','m32','one');
dcl Op3( 3) char init(''  ,'none','imm8','imm32');
dcl Flag(7) char init( '', '', '', 'bito', 'mr', 'cc', 'opd', 'opr' );

MatchOp1: proc( string ) fix;
  dcl (string, t) char;
  dcl i fix;

  do i = 0 to 17;
    t = strtok( Op1(i), ',' );
    do while length(t) > 0;
      if string = t then
        return i;
      t = strtok( NULL, ',' );
    end;
  end;

  return -1;
end MatchOp1;

MatchOp2: proc( string ) fix;
  dcl string char;
  dcl i fix;

  do i = 0 to 15;
    if string = Op2(i) then
      return i;
  end;

  return -1;
end MatchOp2;

MatchOp3: proc( string ) fix;
  dcl string char;
  dcl i fix;

  do i = 0 to 3;
    if string = Op3(i) then
      return i + 4;
  end;

  return -1;
end MatchOp3;

MatchFlag: proc( string ) fix;
  dcl string char;
  dcl i fix;

  do i = 0 to 7;
    if string = Flag(i) then
      return shl( 1, i );
  end;

  return -1;
end MatchFlag;


//
// Add a new INST
//
NewINST: proc( inst ) fix;
  dcl inst char;
  dcl i fix;

  do i = 0 to nInst - 1;
    if inst = INST(i) then
      return i;
  end;

  if nInst > MaxTEMPLATE then
    do;
      Stderr = '*** MaxTEMPLATE too small.';
      call exit;
    end;

  INST(nInst) = inst;
  nInst = nInst + 1;

  return nInst - 1;
end NewINST;


//
// Add a new OPCODES
//
NewOPCODES: proc( opcode ) fix;
  dcl opcode char;
  dcl i fix;

  do i = 0 to nOpcodes - 1;
    if opcode = OPCODES(i) then
      return i;
  end;

  if nOpcodes > MaxOPCODES then
    do;
      Stderr = '*** MaxOPCODES too small.';
      call exit;
    end;

  OPCODES(nOpcodes) = opcode;
  nOpcodes = nOpcodes + 1;

  return nOpcodes - 1;
end NewOPCODES;


//
// Read the instruction table
//
ReadTable: proc;
  dcl start bool;
  dcl (p, v, v1, v2, v3, v4, flag) fix;
  dcl (Line, f1, f2, f3, f4, i1, i2, i3, i4) char;

  start = false;

  do forever;
    Line = Stdin;
    if length(Line) = 0 then
      return;

    if start then
      do;
        if byte(substr(Line,0,1)) ~= byte(';') then
          do;
            p = strchr( Line, byte(';') );
            if p >= 0 then
              Line = substr( Line, 0, p );

            f1 = strtok( Line, ' ' );
            f2 = strtok( NULL, ' ' );
            f3 = strtok( NULL, ' ' );
            f4 = strtok( NULL, ' ' );

            if length(f1) = 0 | length(f2) = 0 | length(f3) = 0 then
              do;
                Stderr = '*** Instruction table line contains too few fields:';
                Stderr = Line;
              end;
            else
              do;
                i1 = strtok( f1, ',' );
                i2 = strtok( NULL, ',' );
                if i2 = '' then
                  i2, i3, i4 = 'none';
                else
                  do;
                    i3 = strtok( NULL, ',' );
                    if i3 = '' then
                      i3, i4 = 'none';
                    else
                      do;
                        i4 = strtok( NULL, ',' );
                        if i4 = '' then
                          i4 = 'none';
                      end;
                  end;

                v1 = NewINST( i1 );
                v2 = MatchOp1( i2 );
                v3 = MatchOp2( i3 );
                v4 = MatchOp3( i4 );

                if v1 = -1 | v2 = -1 | v3 = -1 | v4 = -1 then
                  do;
                    Stderr = '*** Instruction table template contains unknown items:';
                    Stderr = Line;
                  end;
                else
                  do;
                    if nTemplate > MaxTEMPLATE then
                      do;
                        Stderr = '*** MaxTEMPLATE too small.';
                        call exit;
                      end;

                    v = Inst( v1, v2, v3, v4 );

                    TEMPLATE(nTemplate) = 'CCC' || 'XX' || 'F' || 'I';

                    byte( TEMPLATE(nTemplate), 0 ) = shr( v, 16 );
                    byte( TEMPLATE(nTemplate), 1 ) = shr( v, 8 );
                    byte( TEMPLATE(nTemplate), 2 ) =      v;

                    if strchr(f3,byte(',')) = -1 then
                      do;
                        v = atox( substr(f3,1,length(f3)-2) );

                        byte( TEMPLATE(nTemplate), 3 ) = shr( v, 8 );
                        byte( TEMPLATE(nTemplate), 4 ) =      v;
                      end;
                    else
                      do;
                        if nRelopcodes > MaxRELOPCODES then
                          do;
                            Stderr = '*** MaxRELOPCODES too small.';
                            call exit;
                          end;

                        byte( TEMPLATE(nTemplate), 3 ) = 0;
                        byte( TEMPLATE(nTemplate), 4 ) = nRelopcodes;

                        f3 = strtok( f3, ',' );
                        RELOPCODES(nRelopcodes) = atox( substr(f3,1,length(f3)-2) );
                        nRelopcodes = nRelopcodes + 1;

                        f3 = strtok( NULL, ',' );
                        RELOPCODES(nRelopcodes) = atox( substr(f3,1,length(f3)-2) );
                        nRelopcodes = nRelopcodes + 1;
                      end;

                    flag = 0;
                    f4 = strtok( f4, ',' );
                    do while length(f4) > 0;
                      if byte('0') <= byte(f4) & byte(f4) <= byte('9') then
                        flag = flag | atoi( f4 );
                      else
                        do;
                          v = MatchFlag( f4 );
                          if v = -1 then
                            do;
                              Stderr = '*** Instruction table flag contains unknown items:';
                              Stderr = Line;
                            end;
                          else
                            flag = flag | MatchFlag( f4 );
                        end;
                      f4 = strtok( NULL, ',' );
                    end;

                    byte( TEMPLATE(nTemplate), 5 ) = flag;

                    byte( TEMPLATE(nTemplate), 6 ) =
                                        NewOPCODES( substr(f2,1,length(f2)-2) );

                    nTemplate = nTemplate + 1;
                  end;
              end;
          end;
      end;
    else
      if substr(Line,0,3) = 'EOF' then
        start = true;
  end;
end ReadTable;


// Debugging printouts
//
DumpTemplate: proc;
  dcl (i, code, uop, ua1, ua2, ua3, f) fix;
  dcl (t1, t2) char;
  Stdout = '';
  Stdout = 'TEMPLATE';
  do i = 0 to nTemplate - 1;
    code = shl( byte(TEMPLATE(i),0), 16 ) | shl( byte(TEMPLATE(i),1), 8 )
                                                          | byte(TEMPLATE(i),2);
    uop = shr( code, 17 ) & "7F";
    ua1 = shr( code, 9 ) & "FF";
    ua2 = shr( code, 2 ) & "7F";
    ua3 = code & "03";

    t1 = INST(uop) || ',' || strtok(Op1(shr(ua1,3)),',') || ',' ||
                                             Op2(shr(ua2,3)) || ',' || Op3(ua3);
    t2 = '';
    do f = 3 to 7;
      if (byte(TEMPLATE(i),5) & shl(1,f)) ~= 0 then
        do;
          if t2 = '' then
            t2 = Flag(f);
          else
            t2 = t2 || '|' || Flag(f);
        end;
    end;
    if (byte(TEMPLATE(i),5) & "80") ~= 0 then
      t2 = t2 || '|' || (byte(TEMPLATE(i),5) & "07");

    Stdout = 'TEMPLATE(' || i || ')="(4) ' || Xformat(byte(TEMPLATE(i),0),2) || 
                                              Xformat(byte(TEMPLATE(i),1),2) || 
                                              Xformat(byte(TEMPLATE(i),2),2) || ' ' ||
                                              Xformat(byte(TEMPLATE(i),3),2) || 
                                              Xformat(byte(TEMPLATE(i),4),2) || ' ' ||
                                              Xformat(byte(TEMPLATE(i),5),2) || ' ' ||
                                              Xformat(byte(TEMPLATE(i),6),2) || '"'
                                              || '     ' ||
                                              Pad( t1, 23 )
                                              || '     ' ||
                                              Pad( OPCODES( byte(TEMPLATE(i),6) ), 8 )
                                              || '     ' ||
                                              t2;
  end;
end DumpTemplate;

DumpRelopcodes: proc;
  dcl i fix;
  Stdout = '';
  Stdout = 'RELOPCODES';
  do i = 0 to nRelopcodes - 1 BY 2;
    Stdout = '   RELOPCODES(' || i || ',' || (i + 1) || ') = "' ||
             Xformat(RELOPCODES(i),4) || '","' || Xformat(RELOPCODES(i+1),4) || '"';
  end;
end DumpRelopcodes;

DumpOpcodes: proc;
  dcl i fix;
  Stdout = '';
  Stdout = 'Sorted OPCODES';
  do i = 0 to nOpcodes - 1;
    Stdout = '   OPCODES(' || i || ') = ' ||
             Pad( temp(i) || '->' || newx(temp(i)), 8 ) || '  ''' || OPCODES(i) || '''';
  end;
end DumpOpcodes;

DumpInst: proc;
  dcl i fix;
  Stdout = '';
  Stdout = 'Sorted INST';
  do i = 0 to nInst - 1;
    Stdout = '   INST(' || i || ') = ' ||
             Pad( temp(i) || '->' || newx(temp(i)), 8 ) || '  ''' || INST(i) || '''';
  end;
end DumpInst;



// Begin...
//
dcl (i, j, k, l, lim, t, t1, t2) fix;
dcl s char;
dcl debug bool init( false );

// Command line...
//
s = strtok( parmstring, ',' );
do while length(s) > 0;
  if s = 'debug' then
    debug = true;
  s = strtok( NULL, ',' );
end;


// Read the instruction template table...
//
call ReadTable;

if debug then
  do;
    call DumpTemplate;
    call DumpRelopcodes;
  end;


// Alphabetic order is aesthetically pleasing...
//

//   Sort OPCODES
do i = 0 to nOpcodes - 1;
  temp(i) = i;
end;

lim = nOpcodes - 1;
k = shr( lim+1, 1 );
do while k > 0;
  do j = 0 to lim - k;
    i = j;
    do while strcmp( OPCODES(i), OPCODES(i+k) ) > 0;
      l = i + k;
      t = temp(l);     temp(l) = temp(i);         temp(i) = t;
      s = OPCODES(l);  OPCODES(l) = OPCODES(i);   OPCODES(i) = s;
      i = i - k;
      if i < 0 then
        goto X1;
      end;
X1: end;
  k = shr( k, 1 );
end;

do i = 0 to nOpcodes - 1;
  newx( temp(i) ) = i;
end;

do i = 0 to nTemplate - 1;
  byte( TEMPLATE(i), 6 ) = newx( byte(TEMPLATE(i),6) );
end;

if debug then
  do;
    call DumpOpcodes;
    call DumpTemplate;
  end;


//   Sort INST
do i = 0 to nInst - 1;
  temp(i) = i;
end;

lim = nInst - 1;
k = shr( lim+1, 1 );
do while k > 0;
  do j = 0 to lim - k;
    i = j;
    do while strcmp( INST(i), INST(i+k) ) > 0;
      l = i + k;
      t = temp(l);  temp(l) = temp(i);   temp(i) = t;
      s = INST(l);  INST(l) = INST(i);   INST(i) = s;
      i = i - k;
      if i < 0 then
        goto X2;
      end;
X2: end;
  k = shr( k, 1 );
end;

do i = 0 to nInst - 1;
  newx( temp(i) ) = i;
end;

do i = 0 to nTemplate - 1;
  byte( TEMPLATE(i) ) = (byte(TEMPLATE(i)) & "01") |
                                    shl( newx( shr(byte(TEMPLATE(i)),1) ), 1 );
end;

if debug then
  do;
    call DumpInst;
    call DumpTemplate;
  end;


//   Sort TEMPLATE
//
lim = nTemplate - 1;
k = shr( lim+1, 1 );
do while k > 0;
  do j = 0 to lim - k;
    i = j;
    t1 = shl(byte(TEMPLATE(i)),16)   | shl(byte(TEMPLATE(i),1),8)   | byte(TEMPLATE(i),2);
    t2 = shl(byte(TEMPLATE(i+k)),16) | shl(byte(TEMPLATE(i+k),1),8) | byte(TEMPLATE(i+k),2);
    do while t1 > t2;
      l = i + k;
      s = TEMPLATE(l);  TEMPLATE(l) = TEMPLATE(i);   TEMPLATE(i) = s;
      i = i - k;
      if i < 0 then
        goto X3;
      t1 = shl(byte(TEMPLATE(i)),16)   | shl(byte(TEMPLATE(i),1),8)   | byte(TEMPLATE(i),2);
      end;
X3: end;
  k = shr( k, 1 );
end;

if debug then
  call DumpTemplate;


// Write the XPL data statements
//
output(3) = '/* Start of cards punched by TEMPLATE.XPL */';
output(3) = '';

output(3) = 'declare NTEMPLATE literally ''' || (nTemplate - 1) || ''';';

s = 'declare template(NTEMPLATE) bit(56) initial(';
do i = 0 to nTemplate - 1;
  if (i MOD 3) = 0 then
    do;
      output(3) = s;
      s = '         ';
    end;
  else
    s = s || ' ';

  t = shl(byte(TEMPLATE(i),0),16) | shl(byte(TEMPLATE(i),1),8) | byte(TEMPLATE(i),2);
  t1 = shl(byte(TEMPLATE(i),3),8) | byte(TEMPLATE(i),4);

  s = s || '"' || Xformat(t,6) || ' ' ||
                  Xformat(t1,4) || ' ' ||
                  Xformat(byte(TEMPLATE(i),5),2) || ' ' ||
                  Xformat(byte(TEMPLATE(i),6),2) || '"';

  if i ~= (nTemplate - 1) then
    s = s || ',';
end;
output(3) = s || ' );';
output(3) = '';

output(3) = 'declare relopcodes(' || (nRelopcodes - 1) || ') bit(16) initial(';
s = '     "' || Xformat(RELOPCODES(0),4) || '"';
do i = 1 to nRelopcodes - 1;
  s = s || ', "' || Xformat(RELOPCODES(i),4) || '"';
end;
output(3) = s || ' );';
output(3) = '';

output(3) = 'declare NOPCODE literally ''' || (nOpcodes - 1) || ''';';
s = 'declare i486Opcodes(NOPCODE) character initial( ';
do i = 0 to nOpcodes - 1;
  if length(s) > 67 then
    do;
      output(3) = s || ',';
      s = '     ''' || OPCODES(i) || '''';
    end;
  else
    do;
      if i = 0 then
        s = s || '''' || OPCODES(i) || '''';
      else
        s = s || ', ''' || OPCODES(i) || '''';
    end;
end;
output(3) = s || ' );';
output(3) = '';

output(3) = 'declare NINST literally ''' || (nInst - 1) || ''';';
do i = 0 to nInst - 1;
  j = strchr( INST(i), byte('=') );
  if j = -1 then
    output(3) = 'declare ' || Pad(toUpper(INST(i)),6) || ' literally ''' || i || ''';';
  else
    output(3) = 'declare ' || Pad(substr(toUpper(INST(i)),j+1),6) || ' literally '''
                                        || i || ''';        // ' || substr(INST(i),0,j);
end;
output(3) = '';

output(3) = '/* End of cards punched by TEMPLATE.XPL */';

return 0;


EOF EOF EOF
;
; Template                Opcode     Value(s)     Flags
; ----------------------  ---------  -----------  ----------
  aaa                     'aaa'      "37"
  aad                     'aad'      "D50A"
  aam                     'aam'      "D40A"
  aas                     'aas'      "3F"
  adc,al,imm8             'adcb'     "14"
  adc,eax,imm32           'adc'      "15"
  adc,m16,imm16           'adcw'     "6681"       opr,2
  adc,m16,imm8            'adcw'     "6683"       opr,2
  adc,m16,r16             'adcw'     "6611"
  adc,m32,imm32           'adc'      "81"         opr,2
  adc,m32,imm8            'adc'      "83"         opr,2
  adc,m32,r32             'adc'      "11"
  adc,m8,imm8             'adcb'     "80"         opr,2
  adc,m8,r8               'adcb'     "10"
  adc,r16,m16             'adcw'     "6613"       mr
  adc,r32,imm32           'adc'      "81"         opr,2
  adc,r32,imm8            'adc'      "83"         opr,2
  adc,r32,m32             'adc'      "13"         mr
  adc,r32,r32             'adc'      "11"
; adc,r32,r32             'adc'      "13"         mr             ; Duplicated by alternate encoding
  adc,r8,imm8             'adcb'     "80"         opr,2
  adc,r8,m8               'adcb'     "12"         mr
  adc,r8,r8               'adcb'     "10"
; adc,r8,r8               'adcb'     "12"         mr             ; Duplicated by alternate encoding
  add,al,imm8             'addb'     "04"
  add,eax,imm32           'add'      "05"
  add,m16,imm16           'addw'     "6681"       opr,0
  add,m16,imm8            'addw'     "6683"       opr,0
  add,m16,r16             'addw'     "6601"
  add,m32,imm32           'add'      "81"         opr,0
  add,m32,imm8            'add'      "83"         opr,0
  add,m32,r32             'add'      "01"
  add,m8,imm8             'addb'     "80"         opr,0
  add,m8,r8               'addb'     "00"
  add,r16,m16             'addw'     "6603"       mr
  add,r32,imm32           'add'      "81"         opr,0
  add,r32,imm8            'add'      "83"         opr,0
  add,r32,m32             'add'      "03"         mr
  add,r32,r32             'add'      "01"
; add,r32,r32             'add'      "03"         mr             ; Duplicated by alternate encoding
  add,r8,imm8             'addb'     "80"         opr,0
  add,r8,m8               'addb'     "02"         mr
  add,r8,r8               'addb'     "00"
; add,r8,r8               'addb'     "02"         mr             ; Duplicated by alternate encoding
  and,al,imm8             'andb'     "24"
  and,eax,imm32           'and'      "25"
  and,m16,imm16           'andw'     "6681"       opr,4
  and,m16,imm8            'andw'     "6683"       opr,4
  and,m16,r16             'andw'     "6621"
  and,m32,imm32           'and'      "81"         opr,4
  and,m32,imm8            'and'      "83"         opr,4
  and,m32,r32             'and'      "21"
  and,m8,imm8             'andb'     "80"         opr,4
  and,m8,r8               'andb'     "20"
  and,r16,m16             'andw'     "6623"       mr
  and,r32,imm32           'and'      "81"         opr,4
  and,r32,imm8            'and'      "83"         opr,4
  and,r32,m32             'and'      "23"         mr
  and,r32,r32             'and'      "21"
; and,r32,r32             'and'      "23"         mr             ; Duplicated by alternate encoding
  and,r8,imm8             'andb'     "80"         opr,4
  and,r8,m8               'andb'     "22"         mr
  and,r8,r8               'andb'     "20"
; and,r8,r8               'andb'     "22"         mr             ; Duplicated by alternate encoding
  bsf,r32,m32             'bsf'      "0FBC"       mr
  bsf,r32,r32             'bsf'      "0FBC"       mr
  bsr,r32,m32             'bsr'      "0FBD"       mr
  bsr,r32,r32             'bsr'      "0FBD"       mr
  bswap,r32               'bswap'    "0FC8"       opd
  bt,m32,imm8             'bt'       "0FBA"       bito,opr,4
  bt,m32,r32              'bt'       "0FA3"
  bt,r32,imm8             'bt'       "0FBA"       opr,4
  bt,r32,r32              'bt'       "0FA3"
  btc,m32,imm8            'btc'      "0FBA"       bito,opr,7
  btc,m32,r32             'btc'      "0FBB"
  btc,r32,imm8            'btc'      "0FBA"       opr,7
  btc,r32,r32             'btc'      "0FBB"
  btr,m32,imm8            'btr'      "0FBA"       bito,opr,6
  btr,m32,r32             'btr'      "0FB3"
  btr,r32,imm8            'btr'      "0FBA"       opr,6
  btr,r32,r32             'btr'      "0FB3"
  bts,m32,imm8            'bts'      "0FBA"       bito,opr,5
  bts,m32,r32             'bts'      "0FAB"
  bts,r32,imm8            'bts'      "0FBA"       opr,5
  bts,r32,r32             'bts'      "0FAB"
  call=jsr,m32            'call'     "FF"         opr,2
  call=jsr,r32            'call'     "FF"         opr,2
  call=jsr,rel32          'call'     "E8"
  cbw                     'cbw'      "6698"
  cdq                     'cdq'      "99"
  clc                     'clc'      "F8"
  cld                     'cld'      "FC"
  cmc                     'cmc'      "F5"
  cmp,al,imm8             'cmpb'     "3C"                        ; cc <= op1 - op2
  cmp,eax,imm32           'cmp'      "3D"
  cmp,m16,imm16           'cmpw'     "6681"       opr,7
  cmp,m16,imm8            'cmpw'     "6683"       opr,7
  cmp,m16,r16             'cmpw'     "6639"
  cmp,m32,imm32           'cmp'      "81"         opr,7
  cmp,m32,imm8            'cmp'      "83"         opr,7
  cmp,m32,r32             'cmp'      "39"
  cmp,m8,imm8             'cmpb'     "80"         opr,7
  cmp,m8,r8               'cmpb'     "38"
  cmp,r16,m16             'cmpw'     "663B"       mr
  cmp,r32,imm32           'cmp'      "81"         opr,7
  cmp,r32,imm8            'cmp'      "83"         opr,7
  cmp,r32,m32             'cmp'      "3B"         mr
  cmp,r32,r32             'cmp'      "39"
; cmp,r32,r32             'cmp'      "3B"         mr             ; Duplicated by alternate encoding
  cmp,r8,imm8             'cmpb'     "80"         opr,7
  cmp,r8,m8               'cmpb'     "3A"         mr
  cmp,r8,r8               'cmpb'     "38"
; cmp,r8,r8               'cmpb'     "3A"         mr             ; Duplicated by alternate encoding
  cmpsb                   'cmpsb'    "A6"
  cmpsd                   'cmpsd'    "A7"
  cwde                    'cwde'     "98"
  daa                     'daa'      "27"
  das                     'das'      "2F"
  dec,m16                 'decw'     "66FF"       opr,1
  dec,m32                 'dec'      "FF"         opr,1
  dec,m8                  'decb'     "FE"         opr,1
  dec,r32                 'dec'      "48"         opd
; dec,r32                 'dec'      "FF"         opr,1          ; Duplicated by alternate encoding
  dec,r8                  'decb'     "FE"         opr,1
  div,eax,m32             'div'      "F7"         mr,opr,6       ; edx:eax / m32
  div,eax,r32             'div'      "F7"         mr,opr,6       ; edx:eax / r32
  idiv,eax,m32            'idiv'     "F7"         mr,opr,7       ; edx:eax / m32
  idiv,eax,r32            'idiv'     "F7"         mr,opr,7       ; edx:eax / r32
  imul,eax,m32            'imul'     "F7"         mr,opr,5       ; edx:eax = eax * m32
  imul,eax,r32            'imul'     "F7"         mr,opr,5       ; edx:eax = eax * r32
  imul,m32                'imul'     "F7"         opr,5          ; edx:eax = eax * m32
  imul,r32                'imul'     "F7"         opr,5          ; edx:eax = eax * r32
  imul,r32,imm32          'imul'     "69"         mr             ; r32 *= imm32
  imul,r32,imm8           'imul'     "6B"         mr             ; r32 *= imm8
  imul,r32,m32            'imul'     "0FAF"       mr             ; r32 *= m32
  imul,r32,m32,imm32      'imul'     "69"         mr             ; r32 = m32 * imm32
  imul,r32,m32,imm8       'imul'     "6B"         mr             ; r32 = m32 * imm8
  imul,r32,r32            'imul'     "0FAF"       mr             ; r32 *= r32
  imul,r32,r32,imm32      'imul'     "69"         mr             ; r32 = r32 * imm32
  imul,r32,r32,imm8       'imul'     "6B"         mr             ; r32 = r32 * imm8
  inc,m16                 'incw'     "66FF"       opr,0
  inc,m32                 'inc'      "FF"         opr,0
  inc,m8                  'incb'     "FE"         opr,0
  inc,r32                 'inc'      "40"         opd
; inc,r32                 'inc'      "FF"         opr,0          ; Duplicated by alternate encoding
  inc,r8                  'incb'     "FE"         opr,0
  jcc,rel                 'j'        "70","0F80"  cc
  jcc,rel32               'j'        "0F80"       cc
  jcc,rel8                'j'        "70"         cc
  jcxz,rel8               'jcxz'     "E3"
  jmp,m32                 'jmp'      "FF"         opr,4
  jmp,r32                 'jmp'      "FF"         opr,4
  jmp,rel                 'jmp'      "EB","E9"    
  jmp,rel32               'jmp'      "E9"
  jmp,rel8                'jmp'      "EB"
  lahf                    'lahf'     "9F"
  lahf,ah                 'lahf'     "9F"
  lea,r32,m32             'lea'      "8D"         mr
  lodsb                   'lodsb'    "AC"
  lodsd                   'lodsd'    "AD"
  loop,rel8               'loop'     "E2"
  loopnz,rel8             'loopnz'   "E0"
  loopz,rel8              'loopz'    "E1"
  mov,m16,imm16           'movw'     "66C7"       opr,0
  mov,m16,r16             'movw'     "6689"
  mov,m32,imm32           'mov'      "C7"         opr,0
  mov,m32,r32             'mov'      "89"
  mov,m8,imm8             'movb'     "C6"         opr,0
  mov,m8,r8               'movb'     "88"
  mov,r16,m16             'movw'     "668B"       mr
  mov,r32,imm32           'mov'      "B8"         opd
; mov,r32,imm32           'mov'      "C7"         opr,0          ; Duplicated by alternate encoding
  mov,r32,m32             'mov'      "8B"         mr
  mov,r32,r32             'mov'      "89"
; mov,r32,r32             'mov'      "8B"         mr             ; Duplicated by alternate encoding
  mov,r8,imm8             'movb'     "B0"         opd
; mov,r8,imm8             'movb'     "C6"         opr,0          ; Duplicated by alternate encoding
  mov,r8,m8               'movb'     "8A"         mr
  mov,r8,r8               'movb'     "88"
; mov,r8,r8               'movb'     "8A"         mr             ; Duplicated by alternate encoding
  movsb                   'movsb'    "A4"
  movsd                   'movsd'    "A5"
  movsx,r32,m16           'movsxw'   "0FBF"       mr
  movsx,r32,m8            'movsxb'   "0FBE"       mr
  movsx,r32,r16           'movsxw'   "0FBF"       mr
  movsx,r32,r8            'movsxb'   "0FBE"       mr
  movzx,r32,m16           'movzxw'   "0FB7"       mr
  movzx,r32,m8            'movzxb'   "0FB6"       mr
  movzx,r32,r16           'movzxw'   "0FB7"       mr
  movzx,r32,r8            'movzxb'   "0FB6"       mr
  mul,eax,m32             'mul'      "F7"         mr,opr,4       ; edx:eax = eax * m32
  mul,eax,r32             'mul'      "F7"         mr,opr,4       ; edx:eax = eax * r32
  neg,m16                 'negw'     "66F7"       opr,3
  neg,m32                 'neg'      "F7"         opr,3
  neg,m8                  'negb'     "F6"         opr,3
  neg,r32                 'neg'      "F7"         opr,3
  neg,r8                  'negb'     "F6"         opr,3
  nop                     'nop'      "90"
  not,m16                 'notw'     "66F7"       opr,2
  not,m32                 'not'      "F7"         opr,2
  not,m8                  'notb'     "F6"         opr,2
  not,r32                 'not'      "F7"         opr,2
  not,r8                  'notb'     "F6"         opr,2
  or,al,imm8              'orb'      "0C"
  or,eax,imm32            'or'       "0D"
  or,m16,imm16            'orw'      "6681"       opr,1
  or,m16,imm8             'orw'      "6683"       opr,1
  or,m16,r16              'orw'      "6609"
  or,m32,imm32            'or'       "81"         opr,1
  or,m32,imm8             'or'       "83"         opr,1
  or,m32,r32              'or'       "09"
  or,m8,imm8              'orb'      "80"         opr,1
  or,m8,r8                'orb'      "08"
  or,r16,m16              'orw'      "660B"       mr
  or,r32,imm32            'or'       "81"         opr,1
  or,r32,imm8             'or'       "83"         opr,1
  or,r32,m32              'or'       "0B"         mr
  or,r32,r32              'or'       "09"
; or,r32,r32              'or'       "0B"         mr             ; Duplicated by alternate encoding
  or,r8,imm8              'orb'      "80"         opr,1
  or,r8,m8                'orb'      "0A"         mr
  or,r8,r8                'orb'      "08"
; or,r8,r8                'orb'      "0A"         mr             ; Duplicated by alternate encoding
  pop,m32                 'pop'      "8F"         opr,0
  pop,r32                 'pop'      "58"         opd
  popa                    'popa'     "61"
  popf                    'popf'     "9D"
  push,imm32              'push'     "68"
  push,imm8               'push'     "6A"
  push,m32                'push'     "FF"         opr,6
  push,r32                'push'     "50"         opd
  pusha                   'pusha'    "60"
  pushf                   'pushf'    "9C"
  rcl,m16,cl              'rclw'     "66D3"       opr,2
  rcl,m16,imm8            'rclw'     "66C1"       opr,2
  rcl,m16,one             'rclw'     "66D1"       opr,2
  rcl,m32,cl              'rcl'      "D3"         opr,2
  rcl,m32,imm8            'rcl'      "C1"         opr,2
  rcl,m32,one             'rcl'      "D1"         opr,2
  rcl,m8,cl               'rclb'     "D2"         opr,2
  rcl,m8,imm8             'rclb'     "C0"         opr,2
  rcl,m8,one              'rclb'     "D0"         opr,2
  rcl,r32,cl              'rcl'      "D3"         opr,2
  rcl,r32,imm8            'rcl'      "C1"         opr,2
  rcl,r32,one             'rcl'      "D1"         opr,2
  rcl,r8,cl               'rclb'     "D2"         opr,2
  rcl,r8,imm8             'rclb'     "C0"         opr,2
  rcl,r8,one              'rclb'     "D0"         opr,2
  rcr,m16,cl              'rcrw'     "66D3"       opr,3
  rcr,m16,imm8            'rcrw'     "66C1"       opr,3
  rcr,m16,one             'rcrw'     "66D1"       opr,3
  rcr,m32,cl              'rcr'      "D3"         opr,3
  rcr,m32,imm8            'rcr'      "C1"         opr,3
  rcr,m32,one             'rcr'      "D1"         opr,3
  rcr,m8,cl               'rcrb'     "D2"         opr,3
  rcr,m8,imm8             'rcrb'     "C0"         opr,3
  rcr,m8,one              'rcrb'     "D0"         opr,3
  rcr,r32,cl              'rcr'      "D3"         opr,3
  rcr,r32,imm8            'rcr'      "C1"         opr,3
  rcr,r32,one             'rcr'      "D1"         opr,3
  rcr,r8,cl               'rcrb'     "D2"         opr,3
  rcr,r8,imm8             'rcrb'     "C0"         opr,3
  rcr,r8,one              'rcrb'     "D0"         opr,3
  rep                     'rep'      "F3"
  repnz                   'repnz'    "F2"
  ret                     'ret'      "C3"
  ret,imm16               'ret'      "C2"
  rol,m16,cl              'rolw'     "66D3"       opr,0
  rol,m16,imm8            'rolw'     "66C1"       opr,0
  rol,m16,one             'rolw'     "66D1"       opr,0
  rol,m32,cl              'rol'      "D3"         opr,0
  rol,m32,imm8            'rol'      "C1"         opr,0
  rol,m32,one             'rol'      "D1"         opr,0
  rol,m8,cl               'rolb'     "D2"         opr,0
  rol,m8,imm8             'rolb'     "C0"         opr,0
  rol,m8,one              'rolb'     "D0"         opr,0
  rol,r32,cl              'rol'      "D3"         opr,0
  rol,r32,imm8            'rol'      "C1"         opr,0
  rol,r32,one             'rol'      "D1"         opr,0
  rol,r8,cl               'rolb'     "D2"         opr,0
  rol,r8,imm8             'rolb'     "C0"         opr,0
  rol,r8,one              'rolb'     "D0"         opr,0
  ror,m16,cl              'rorw'     "66D3"       opr,1
  ror,m16,imm8            'rorw'     "66C1"       opr,1
  ror,m16,one             'rorw'     "66D1"       opr,1
  ror,m32,cl              'ror'      "D3"         opr,1
  ror,m32,imm8            'ror'      "C1"         opr,1
  ror,m32,one             'ror'      "D1"         opr,1
  ror,m8,cl               'rorb'     "D2"         opr,1
  ror,m8,imm8             'rorb'     "C0"         opr,1
  ror,m8,one              'rorb'     "D0"         opr,1
  ror,r32,cl              'ror'      "D3"         opr,1
  ror,r32,imm8            'ror'      "C1"         opr,1
  ror,r32,one             'ror'      "D1"         opr,1
  ror,r8,cl               'rorb'     "D2"         opr,1
  ror,r8,imm8             'rorb'     "C0"         opr,1
  ror,r8,one              'rorb'     "D0"         opr,1
  sahf                    'sahf'     "9E"
  sar,m16,cl              'sarw'     "66D3"       opr,7
  sar,m16,imm8            'sarw'     "66C1"       opr,7
  sar,m16,one             'sarw'     "66D1"       opr,7
  sar,m32,cl              'sar'      "D3"         opr,7
  sar,m32,imm8            'sar'      "C1"         opr,7
  sar,m32,one             'sar'      "D1"         opr,7
  sar,m8,cl               'sarb'     "D2"         opr,7
  sar,m8,imm8             'sarb'     "C0"         opr,7
  sar,m8,one              'sarb'     "D0"         opr,7
  sar,r32,cl              'sar'      "D3"         opr,7
  sar,r32,imm8            'sar'      "C1"         opr,7
  sar,r32,one             'sar'      "D1"         opr,7
  sar,r8,cl               'sarb'     "D2"         opr,7
  sar,r8,imm8             'sarb'     "C0"         opr,7
  sar,r8,one              'sarb'     "D0"         opr,7
  sbb,al,imm8             'sbbb'     "1C"
  sbb,eax,imm32           'sbb'      "1D"
  sbb,m16,imm16           'sbbw'     "6681"       opr,3
  sbb,m16,imm8            'sbbw'     "6683"       opr,3
  sbb,m16,r16             'sbbw'     "6619"
  sbb,m32,imm32           'sbb'      "81"         opr,3
  sbb,m32,imm8            'sbb'      "83"         opr,3
  sbb,m32,r32             'sbb'      "19"
  sbb,m8,imm8             'sbbb'     "80"         opr,3
  sbb,m8,r8               'sbbb'     "18"
  sbb,r16,m16             'sbbw'     "661B"       mr
  sbb,r32,imm32           'sbb'      "81"         opr,3
  sbb,r32,imm8            'sbb'      "83"         opr,3
  sbb,r32,m32             'sbb'      "1B"         mr
  sbb,r32,r32             'sbb'      "19"
; sbb,r32,r32             'sbb'      "1B"         mr             ; Duplicated by alternate encoding
  sbb,r8,imm8             'sbbb'     "80"         opr,3
  sbb,r8,m8               'sbbb'     "1A"         mr
  sbb,r8,r8               'sbbb'     "18"
; sbb,r8,r8               'sbbb'     "1A"         mr             ; Duplicated by alternate encoding
  scasb                   'scasb'    "AE"
  scasd                   'scasd'    "AF"
  setcc,m8                'set'      "0F90"       cc
  setcc,r8                'set'      "0F90"       cc
  shl=sll,m16,cl          'shlw'     "66D3"       opr,4
  shl=sll,m16,imm8        'shlw'     "66C1"       opr,4
  shl=sll,m16,one         'shlw'     "66D1"       opr,4
  shl=sll,m32,cl          'shl'      "D3"         opr,4
  shl=sll,m32,imm8        'shl'      "C1"         opr,4
  shl=sll,m32,one         'shl'      "D1"         opr,4
  shl=sll,m8,cl           'shlb'     "D2"         opr,4
  shl=sll,m8,imm8         'shlb'     "C0"         opr,4
  shl=sll,m8,one          'shlb'     "D0"         opr,4
  shl=sll,r32,cl          'shl'      "D3"         opr,4
  shl=sll,r32,imm8        'shl'      "C1"         opr,4
  shl=sll,r32,one         'shl'      "D1"         opr,4
  shl=sll,r8,cl           'shlb'     "D2"         opr,4
  shl=sll,r8,imm8         'shlb'     "C0"         opr,4
  shl=sll,r8,one          'shlb'     "D0"         opr,4
  shr=srl,m16,cl          'shrw'     "66D3"       opr,5
  shr=srl,m16,imm8        'shrw'     "66C1"       opr,5
  shr=srl,m16,one         'shrw'     "66D1"       opr,5
  shr=srl,m32,cl          'shr'      "D3"         opr,5
  shr=srl,m32,imm8        'shr'      "C1"         opr,5
  shr=srl,m32,one         'shr'      "D1"         opr,5
  shr=srl,m8,cl           'shrb'     "D2"         opr,5
  shr=srl,m8,imm8         'shrb'     "C0"         opr,5
  shr=srl,m8,one          'shrb'     "D0"         opr,5
  shr=srl,r32,cl          'shr'      "D3"         opr,5
  shr=srl,r32,imm8        'shr'      "C1"         opr,5
  shr=srl,r32,one         'shr'      "D1"         opr,5
  shr=srl,r8,cl           'shrb'     "D2"         opr,5
  shr=srl,r8,imm8         'shrb'     "C0"         opr,5
  shr=srl,r8,one          'shrb'     "D0"         opr,5
  stc                     'stc'      "F9"
  std                     'std'      "FD"
  stosb                   'stosb'    "AA"
  stosd                   'stosd'    "AB"
  sub,al,imm8             'subb'     "2C"
  sub,eax,imm32           'sub'      "2D"
  sub,m16,imm16           'subw'     "6681"       opr,5
  sub,m16,imm8            'subw'     "6683"       opr,5
  sub,m16,r16             'subw'     "6629"
  sub,m32,imm32           'sub'      "81"         opr,5
  sub,m32,imm8            'sub'      "83"         opr,5
  sub,m32,r32             'sub'      "29"
  sub,m8,imm8             'subb'     "80"         opr,5
  sub,m8,r8               'subb'     "28"
  sub,r16,m16             'subw'     "662B"       mr
  sub,r32,imm32           'sub'      "81"         opr,5
  sub,r32,imm8            'sub'      "83"         opr,5
  sub,r32,m32             'sub'      "2B"         mr
  sub,r32,r32             'sub'      "29"
; sub,r32,r32             'sub'      "2B"         mr             ; Duplicated by alternate encoding
  sub,r8,imm8             'subb'     "80"         opr,5
  sub,r8,m8               'subb'     "2A"         mr
  sub,r8,r8               'subb'     "28"
; sub,r8,r8               'subb'     "2A"         mr             ; Duplicated by alternate encoding
  test,al,imm8            'testb'    "A8"
  test,eax,imm32          'test'     "A9"
  test,m16,imm16          'testw'    "66F7"       opr,0
  test,m16,r16            'testw'    "6685"
  test,m32,imm32          'test'     "F7"         opr,0
  test,m32,r32            'test'     "85"
  test,m8,imm8            'testb'    "F6"         opr,0
  test,m8,r8              'testb'    "84"
  test,r32,imm32          'test'     "F7"         opr,0
  test,r32,r32            'test'     "85"
  test,r8,imm8            'testb'    "F6"         opr,0
  test,r8,r8              'testb'    "84"
  xchg,eax,r32            'xchg'     "90"         opd
  xchg,m16,r16            'xchgw'    "6687"
  xchg,m32,r32            'xchg'     "87"
  xchg,m8,r8              'xchgb'    "86"
  xchg,r16,m16            'xchgw'    "6687"       mr
  xchg,r32,eax            'xchg'     "90"         opd
  xchg,r32,m32            'xchg'     "87"         mr
  xchg,r32,r32            'xchg'     "87"
; xchg,r32,r32            'xchg'     "87"         mr             ; Duplicated by alternate encoding
  xchg,r8,m8              'xchgb'    "86"         mr
  xchg,r8,r8              'xchgb'    "86"
; xchg,r8,r8              'xchgb'    "86"         mr             ; Duplicated by alternate encoding
  xlat                    'xlat'     "D7"
  xlat,al,ebx             'xlat'     "D7"
  xor,al,imm8             'xorb'     "34"
  xor,eax,imm32           'xor'      "35"
  xor,m16,imm16           'xorw'     "6681"       opr,6
  xor,m16,imm8            'xorw'     "6683"       opr,6
  xor,m16,r16             'xorw'     "6631"
  xor,m32,imm32           'xor'      "81"         opr,6
  xor,m32,imm8            'xor'      "83"         opr,6
  xor,m32,r32             'xor'      "31"
  xor,m8,imm8             'xorb'     "80"         opr,6
  xor,m8,r8               'xorb'     "30"
  xor,r16,m16             'xorw'     "6633"       mr
  xor,r32,imm32           'xor'      "81"         opr,6
  xor,r32,imm8            'xor'      "83"         opr,6
  xor,r32,m32             'xor'      "33"         mr
  xor,r32,r32             'xor'      "31"
; xor,r32,r32             'xor'      "33"         mr             ; Duplicated by alternate encoding
  xor,r8,imm8             'xorb'     "80"         opr,6
  xor,r8,m8               'xorb'     "32"         mr
  xor,r8,r8               'xorb'     "30"
; xor,r8,r8               'xorb'     "32"         mr             ; Duplicated by alternate encoding
;
; /* vi: set sw=2 ai sm expandtab: */
