/*
 * String conversions
 *
 * Copyright 1996 David E. Bodnestab
 * All rights reserved
 */

#include <stdlib.h>
#include <ctype.h>


/*
 * ctomem( MEMORY, C-STRING )
 *
 * Convert C-STRING, which is encoded using the usual C syntax for character
 * strings, storing the result in MEMORY.  The length of the result is returned.
 * Any error in the C-STRING is indicated by returning -1.
 */
 int
ctomem( unsigned char * const memory, const unsigned char *cstring )
{
  unsigned char *end = memory;
  unsigned c, n;

  while( (c = *cstring++) != '\0' )
    {
      switch( c )
        {
      /*
       * \?
       */
      case '\\':
	  switch( c = *cstring++ )
	    {
	  default:
	      return -1;
	  case 'a':
	      c = '\07';
	      break;
	  case 'b':
	      c = '\b';
	      break;
	  case 't':
	      c = '\t';
	      break;
	  case 'n':
	      c = '\n';
	      break;
	  case 'v':
	      c = '\v';
	      break;
	  case 'f':
	      c = '\f';
	      break;
	  case 'r':
	      c = '\r';
	      break;
	  case '?':
	      c = '?';
	      break;
	  case '\'':
	      c = '\'';
	      break;
	  case '"':
	      c = '"';
	      break;
	  case '\\':
	      c = '\\';
	      break;
	  /*
	   * \0?, octal, up to three digits
	   */
	  case '0':
	  case '1':
	  case '2':
	  case '3':
	  case '4':
	  case '5':
	  case '6':
	  case '7':
	      n = c - '0';					/* digit 1 */

	      if ( '0' <= (c = *cstring) && c <= '7' )
		{
		  n <<= 3;
		  n |= c - '0';					/* digit 2 */

		  if ( '0' <= (c = *++cstring) && c <= '7' )
		    {
		      n <<= 3;
		      n |= c - '0';				/* digit 3 */

		      ++cstring;
		    }
		}

	      c = n;
	      break;
	  /*
	   * \x?, hexidecimal, unlimited number of hex digits
	   */
	  case 'x':
	  case 'X':
	      if ( (c = *cstring++) == '\0' || ! isxdigit(c) )
	        return -1;

	      n = isdigit(c)? c - '0'				/* digit 1 */
	      		    : isupper(c)? c - 'A' + 10
			    		: c - 'a' + 10;

	      while( (c = *cstring) != '\0' && isxdigit(c) )
	        {
		  n <<= 4;
		  n |= isdigit(c)? c - '0'			/* digit 2,3,4... */
				 : isupper(c)? c - 'A' + 10
					     : c - 'a' + 10;
		  ++cstring;
		}

	      c = n;
	      break;
	    }

      default:
          *end++ = c;
	  break;
	}
    }

  return end - memory;
}


/*
 * memtoc( C-STRING, MEMORY, LENGTH )
 *
 * Convert LENGTH bytes of MEMORY to a character string using the usual C syntax
 * for character strings, storing the result in C-STRING. The length of the
 * result (including the terminating NUL) is returned.
 */
 int
memtoc( unsigned char * const cstring, const unsigned char *memory, unsigned length )
{
  unsigned char *end = cstring;
  unsigned c;

  while( length-- > 0 )
    {
      switch( c = *memory++ )
        {
      default:
	  *end++ = '\\';
	  *end++ = '0' | ((c >> 6) & 07);
	  *end++ = '0' | ((c >> 3) & 07);
	  *end++ = '0' | ((c >> 0) & 07);
          break;
      case '\07':
	  *end++ = '\\';
	  *end++ = 'a';
          break;
      case '\b':
	  *end++ = '\\';
	  *end++ = 'b';
          break;
      case '\t':
	  *end++ = '\\';
	  *end++ = 't';
          break;
      case '\n':
	  *end++ = '\\';
	  *end++ = 'n';
          break;
      case '\v':
	  *end++ = '\\';
	  *end++ = 'v';
          break;
      case '\f':
	  *end++ = '\\';
	  *end++ = 'f';
          break;
      case '\r':
	  *end++ = '\\';
	  *end++ = 'r';
          break;
      case '\'':
	  *end++ = '\\';
	  *end++ = '\'';
          break;
      case '"':
	  *end++ = '\\';
	  *end++ = '"';
          break;
      case '\\':
	  *end++ = '\\';
	  *end++ = '\\';
          break;
      case '?':
	  *end++ = '\\';
	  *end++ = '?';
          break;
      case ' ':
      case '!':
      case '#':
      case '$':
      case '%':
      case '&':
      case '(':
      case ')':
      case '*':
      case '+':
      case ',':
      case '-':
      case '.':
      case '/':
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
      case ':':
      case ';':
      case '<':
      case '=':
      case '>':
      case '@':
      case 'A':
      case 'B':
      case 'C':
      case 'D':
      case 'E':
      case 'F':
      case 'G':
      case 'H':
      case 'I':
      case 'J':
      case 'K':
      case 'L':
      case 'M':
      case 'N':
      case 'O':
      case 'P':
      case 'Q':
      case 'R':
      case 'S':
      case 'T':
      case 'U':
      case 'V':
      case 'W':
      case 'X':
      case 'Y':
      case 'Z':
      case '[':
      case ']':
      case '^':
      case '_':
      case '`':
      case 'a':
      case 'b':
      case 'c':
      case 'd':
      case 'e':
      case 'f':
      case 'g':
      case 'h':
      case 'i':
      case 'j':
      case 'k':
      case 'l':
      case 'm':
      case 'n':
      case 'o':
      case 'p':
      case 'q':
      case 'r':
      case 's':
      case 't':
      case 'u':
      case 'v':
      case 'w':
      case 'x':
      case 'y':
      case 'z':
      case '{':
      case '|':
      case '}':
      case '~':
	  *end++ = c;
          break;
	}
    }

  *end++ = '\0';

  return end - cstring;
}


#ifdef TEST
#include <stdio.h>
#include <memory.h>

unsigned char binary[] = "start \0\01\002\0034\a\b\t\n\v\f\r\'\"\?\\\x1\x02\x003\x0004\777\xfff stop";
unsigned char ascii[] =  "start \\000\\001\\002\\0034\\a\\b\\t\\n\\v\\f\\r\\'\\\"\\?\\\\\\001\\002\\003\\004\\377\\377 stop";

main()
{
  unsigned char buffer[256], see[256];
  int len;

  len = memtoc( buffer, binary, sizeof(binary) - 1 );
  if ( len != sizeof(ascii) )
    fprintf( stderr, "memtoc(binary) returned the incorrect length: %d, not %d\n", len, sizeof(ascii) );
  if ( memcmp(buffer,ascii,len) != 0 )
    fprintf( stderr, "memtoc(binary) returned the incorrect result: \"%s\"\n", buffer );

  len = ctomem( buffer, ascii );
  if ( len != sizeof(binary) - 1 )
    fprintf( stderr, "ctomem(binary) returned the incorrect length: %d, not %d\n", len, sizeof(binary) - 1 );
  if ( memcmp(buffer,binary,len) != 0 )
    {
      memtoc( see, buffer, len );
      fprintf( stderr, "ctomem(binary) returned the incorrect result:\n\t\"%s\", not\n\t\"%s\"\n", buffer, see );
    }
}
#endif


#ifdef DEBUG
#include <stdio.h>
#include <string.h>

main()
{
  unsigned char *p, ascii[256], binary[256], ASCII[256];
  int c, len, i, col;

  while( fputs("\n> ",stdout), fgets(ascii,sizeof(ascii),stdin) != NULL )
    {
      if ( (p = strchr(ascii,'\n')) != NULL )
        *p = '\0';

      len = ctomem( binary, ascii );

      if ( len == -1 )
        {
	  printf( "ctomem(\"%s\") returned -1\n", ascii );
	  continue;
	}

      printf( "result:" );
      for( col = 7, i = 0; i < len; ++i )
        {
	  c = binary[i];

	  if ( c == ' ' || isgraph(c) )
	    {
	      printf( " '%c'", c );
	      col += 4;
	    }
	  else
	    if ( c == '\a' || c == '\b' || c == '\t' || c == '\n' || c == '\v' || c == '\f' || c == '\r' )
	      {
		printf( " \\%c", "abtnvfr"[c-07] );
		col += 3;
	      }
	    else
	      {
		printf( " %03o", c );
		col += 4;
	      }

	  if ( col > 70 && i < len - 1 )
	    {
	      printf( "\n       " );
	      col = 7;
	    }
	}
      putchar( '\n' );

      len = memtoc( ASCII, binary, len );

      printf( "reconverted: \"%s\"\n", ASCII );
    }
}
#endif
