/*
 * Emulator for iAPX486
 *
 * Copyright 1996, David E. Bodenstab
 * All rights reserved.
 *
 * This emulator will successfully trace xcom compiling a small XPL program; it
 * is obviously very slow.  Many instructions are either omitted or incompletely
 * emulated.  The REP,REPNE and OPERAND SIZE instruction prefixes are partially
 * supported.  Only the instructions known to be generated by xcom are fully
 * implemented.
 *
 * Feel free to add support for the unimplemented instructions.
 */

#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <machine/frame.h>
#include "iAPX486.h"


char *regname8[8] = { "al", "cl", "dl", "bl", "ah", "ch", "dh", "bh" };
char *regname16[8] = { "ax", "cx", "dx", "bx", "sp", "bp", "si", "di" };
char *regname32[8] = { "eax", "ecx", "edx", "ebx", "esp", "ebp", "esi", "edi" };

struct modrm
  {
    unsigned char rm : 3;
    unsigned char reg : 3;
    unsigned char mod : 2;
  };

struct sib
  {
    unsigned char base : 3;
    unsigned char index : 3;
    unsigned char ss : 2;
  };

struct instructions
  {
    unsigned flags;
    union
      {
	struct instruction
	  {
	    int code;
	    char *opcode;
	  }
	    instruction;
        struct instructions *next;
      }
        u;
  };

/*
 * Instruction flags
 */
#define	VALID	0x00000001
#define	TWOBYTE	0x00000002	/* Two byte opcode */
#define	GROUPR	0x00000004	/* modrm.reg further identifies the instruction */
#define	REL8	0x00000008	/* call/jmp/jcc rel8 */
#define	REL32	0x00000010	/* call/jmp/jcc rel32 */
#define	IMM16	0x00000020	/* ret n */
#define	DEFOP	0x00000040	/* print opcode after decoding operands */
#define	NOV	0x00000080	/* modrm() should not fetch data values */
#define	MODRM	0x00010000	/* uses mod r/m, dest,src */
#define	OPD	0x00020000	/* opcode + r */
#define	IMM	0x00040000	/* immediate */
#define	MR	0x00080000	/* mod r/m specifies src,dest */
#define	BYTE	0x00100000	/* operand size is byte */
#define	WORD	0x00200000	/* operand size is word (two bytes) */
#define	IMM8	0x00400000	/* immediate byte as long operand (four bytes) */

/*
 * Instruction prefix
 */
#define	REP	0x00000002
#define	REPNE	0x00000004


/*
 * The instructions; an unimplemented instruction is indicated by the *X*.
 */
#define	aaa	0		/*X*/
#define	aad	(aaa+1)		/*X*/
#define	aam	(aad+1)		/*X*/
#define	aas	(aam+1)		/*X*/
#define	adc	(aas+1)
#define	add	(adc+1)
#define	and	(add+1)
#define	bsf	(and+1)		/*X*/
#define	bsr	(bsf+1)		/*X*/
#define	bswap	(bsr+1)
#define	bt	(bswap+1)
#define	btc	(bt+1)
#define	btr	(btc+1)
#define	bts	(btr+1)
#define	call	(bts+1)
#define	cbw	(call+1)	/*X*/
#define	cwde	(cbw+1)
#define	cdq	(cwde+1)
#define	clc	(cdq+1)
#define	cld	(clc+1)
#define	cmc	(cld+1)
#define	cmp	(cmc+1)
#define	cmpsb	(cmp+1)
#define	cmpsd	(cmpsb+1)	/*X*/
#define	cwd	(cmpsd+1)	/*X*/
#define	daa	(cwd+1)		/*X*/
#define	das	(daa+1)		/*X*/
#define	dec	(das+1)
#define	div	(dec+1)		/*X*/
#define	idiv	(div+1)
#define	imul	(idiv+1)
#define	inc	(imul+1)
#define	jcc	(inc+1)
#define	jcxz	(jcc+1)
#define	jmp	(jcxz+1)
#define	lahf	(jmp+1)		/*X*/
#define	lea	(lahf+1)
#define	lodsb	(lea+1)		/*X*/
#define	lodsd	(lodsb+1)	/*X*/
#define	loop	(lodsd+1)	/*X*/
#define	loope	(loop+1)	/*X*/
#define	loopne	(loope+1)	/*X*/
#define	loopnz	loopne		/*X*/
#define	loopz	loope		/*X*/
#define	mov	(loopne+1)
#define	movsb	(mov+1)
#define	movsd	(movsb+1)	/*X*/
#define	movsx	(movsd+1)
#define	movzx	(movsx+1)
#define	mul	(movzx+1)	/*X*/
#define	neg	(mul+1)
#define	nop	(neg+1)		/*X*/
#define	not	(nop+1)
#define	or	(not+1)
#define	pop	(or+1)
#define	popa	(pop+1)		/*X*/
#define	popf	(popa+1)	/*X*/
#define	push	(popf+1)
#define	pusha	(push+1)	/*X*/
#define	pushf	(pusha+1)	/*X*/
#define	rcl	(pushf+1)
#define	rcr	(rcl+1)
#define	rep	(rcr+1)
#define	repe	rep
#define	repne	(rep+1)
#define	repnz	repne
#define	repz	repe
#define	ret	(repne+1)
#define	rol	(ret+1)
#define	ror	(rol+1)
#define	sahf	(ror+1)		/*X*/
#define	sal	(sahf+1)
#define	sar	(sal+1)
#define	sbb	(sar+1)
#define	scasb	(sbb+1)
#define	scasd	(scasb+1)	/*X*/
#define	setcc	(scasd+1)
#define	shl	sal
#define	shr	(setcc+1)
#define	stc	(shr+1)
#define	std	(stc+1)
#define	stosb	(std+1)		/*X*/
#define	stosd	(stosb+1)	/*X*/
#define	sub	(stosd+1)
#define	test	(sub+1)
#define	xchg	(test+1)
#define	xlat	(xchg+1)	/*X*/
#define	xor	(xlat+1)

struct instructions instructionsgroup1[8] ={
	    { VALID | MODRM, { { add, "add" } } },
	    { VALID | MODRM, { { or, "or" } } },
	    { VALID | MODRM, { { adc, "adc" } } },
	    { VALID | MODRM, { { sbb, "sbb" } } },
	    { VALID | MODRM, { { and, "and" } } },
	    { VALID | MODRM, { { sub, "sub" } } },
	    { VALID | MODRM, { { xor, "xor" } } },
	    { VALID | MODRM, { { cmp, "cmp" } } },
				       };

struct instructions instructionsgroup2[8] ={
	    { VALID, { { rol, "rol" } } },
	    { VALID, { { ror, "ror" } } },
	    { VALID, { { rcl, "rcl" } } },
	    { VALID, { { rcr, "rcr" } } },
	    { VALID, { { shl, "shl" } } },
	    { VALID, { { shr, "shr" } } },
	    { VALID, { { shl, "shl" } } },
	    { VALID, { { sar, "sar" } } },
				       };

struct instructions instructionsgroup3[8] ={
	    { VALID | MODRM, { { test, "test" } } },
	    { VALID | MODRM, { { test, "test" } } },
	    { VALID | MODRM, { { not, "not" } } },
	    { VALID | MODRM, { { neg, "neg" } } },
	    { VALID, { { mul, "mul" } } },
	    { VALID | MR, { { imul, "imul" } } },
	    { VALID, { { div, "div" } } },
	    { VALID | MR | MODRM, { { idiv, "idiv" } } },
				       };

struct instructions instructionsgroup4[8] ={
	    { VALID | MODRM | BYTE, { { inc, "inc" } } },
	    { VALID | MODRM | BYTE, { { dec, "dec" } } },
	    { 0 },
	    { 0 },
	    { 0 },
	    { 0 },
	    { 0 },
	    { 0 },
				       };

struct instructions instructionsgroup5[8] ={
	    { VALID | MODRM, { { inc, "inc" } } },
	    { VALID | MODRM, { { dec, "dec" } } },
	    { VALID | MODRM, { { call, "call" } } },
	    { 0 },
	    { VALID | MODRM, { { jmp, "jmp" } } },
	    { 0 },
	    { VALID | MODRM, { { push, "push" } } },
	    { 0 },
				       };

struct instructions instructionsgroup8[8] ={
	    { 0 },
	    { 0 },
	    { 0 },
	    { 0 },
	    { VALID | MODRM | IMM8, { { bt, "bt" } } },
	    { VALID | MODRM | IMM8, { { bts, "bts" } } },
	    { VALID | MODRM | IMM8, { { btr, "btr" } } },
	    { VALID | MODRM | IMM8, { { btc, "btc" } } },
				       };

/*
 * Instructions which have 0x0f as the first byte of the opcode
 */
struct instructions instructions0F[256] ={
  /*00-7F*/ { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
  /* 80 */  { VALID | REL32, { { jcc, "jo" } } },
  /* 81 */  { VALID | REL32, { { jcc, "jno" } } },
  /* 82 */  { VALID | REL32, { { jcc, "jb" } } },
  /* 83 */  { VALID | REL32, { { jcc, "jae" } } },
  /* 84 */  { VALID | REL32, { { jcc, "jz" } } },
  /* 85 */  { VALID | REL32, { { jcc, "jnz" } } },
  /* 86 */  { VALID | REL32, { { jcc, "jbe" } } },
  /* 87 */  { VALID | REL32, { { jcc, "ja" } } },
  /* 88 */  { VALID | REL32, { { jcc, "js" } } },
  /* 89 */  { VALID | REL32, { { jcc, "jns" } } },
  /* 8a */  { VALID | REL32, { { jcc, "jpe" } } },
  /* 8b */  { VALID | REL32, { { jcc, "jpo" } } },
  /* 8c */  { VALID | REL32, { { jcc, "jl" } } },
  /* 8d */  { VALID | REL32, { { jcc, "jge" } } },
  /* 8e */  { VALID | REL32, { { jcc, "jle" } } },
  /* 8f */  { VALID | REL32, { { jcc, "jg" } } },
  /* 90 */  { VALID | MODRM, { { setcc, "seto" } } }, 
  /* 91 */  { VALID | MODRM, { { setcc, "setno" } } },
  /* 92 */  { VALID | MODRM, { { setcc, "setb" } } }, 
  /* 93 */  { VALID | MODRM, { { setcc, "setae" } } },
  /* 94 */  { VALID | MODRM, { { setcc, "setz" } } }, 
  /* 95 */  { VALID | MODRM, { { setcc, "setnz" } } },
  /* 96 */  { VALID | MODRM, { { setcc, "setbe" } } },
  /* 97 */  { VALID | MODRM, { { setcc, "seta" } } }, 
  /* 98 */  { VALID | MODRM, { { setcc, "sets" } } }, 
  /* 99 */  { VALID | MODRM, { { setcc, "setns" } } },
  /* 9a */  { VALID | MODRM, { { setcc, "setpe" } } },
  /* 9b */  { VALID | MODRM, { { setcc, "setpo" } } },
  /* 9c */  { VALID | MODRM, { { setcc, "setl" } } }, 
  /* 9d */  { VALID | MODRM, { { setcc, "setge" } } },
  /* 9e */  { VALID | MODRM, { { setcc, "setle" } } },
  /* 9f */  { VALID | MODRM, { { setcc, "setg" } } }, 
  /* a0 */  { 0 },
  /* a1 */  { 0 },
  /* a2 */  { 0 },
  /* a3 */  { VALID | MODRM, { { bt, "bt" } } },
  /* a4 */  { 0 },
  /* a5 */  { 0 },
  /* a6 */  { 0 },
  /* a7 */  { 0 },
  /* a8 */  { 0 },
  /* a9 */  { 0 },
  /* aa */  { 0 },
  /* ab */  { VALID | MODRM, { { bts, "bts" } } },
  /* ac */  { 0 },
  /* ad */  { 0 },
  /* ae */  { 0 },
  /* af */  { VALID | MR | MODRM, { { imul, "imul" } } },
  /* b0 */  { 0 },
  /* b1 */  { 0 },
  /* b2 */  { 0 },
  /* b3 */  { VALID | MODRM, { { btr, "btr" } } },
  /* b4 */  { 0 },
  /* b5 */  { 0 },
  /* b6 */  { VALID | MODRM | BYTE, { { movzx, "movzx" } } },
  /* b7 */  { VALID | MODRM | WORD, { { movzx, "movzx" } } },
  /* b8 */  { 0 },
  /* b9 */  { 0 },
  /* ba */  { VALID | GROUPR, { next: instructionsgroup8 } },
  /* bb */  { VALID | MODRM, { { btc, "btc" } } },
  /* bc */  { VALID, { { bsf, "bsf" } } },
  /* bd */  { VALID, { { bsr, "bsr" } } },
  /* be */  { VALID | MODRM | BYTE, { { movsx, "movsx" } } },
  /* bf */  { VALID | MODRM | WORD, { { movsx, "movsx" } } },
  /* c0 */  { 0 },
  /* c1 */  { 0 },
  /* c2 */  { 0 },
  /* c3 */  { 0 },
  /* c4 */  { 0 },
  /* c5 */  { 0 },
  /* c6 */  { 0 },
  /* c7 */  { 0 },
  /* c8 */  { VALID | OPD, { { bswap, "bswap" } } },
  /* c9 */  { VALID | OPD, { { bswap, "bswap" } } },
  /* ca */  { VALID | OPD, { { bswap, "bswap" } } },
  /* cb */  { VALID | OPD, { { bswap, "bswap" } } },
  /* cc */  { VALID | OPD, { { bswap, "bswap" } } },
  /* cd */  { VALID | OPD, { { bswap, "bswap" } } },
  /* ce */  { VALID | OPD, { { bswap, "bswap" } } },
  /* cf */  { VALID | OPD, { { bswap, "bswap" } } },
  /*d0-ff*/ { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	    { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
				       };

/*
 * Instructions which have 0x66 (operand size prefix) as the first byte of the
 * opcode
 *
 * This table is incomplete.  It only supports instructions known to be
 * generated by xcom.  For instance, all instructions (such as ADD) do
 * not appear for the form 'OPSIZ INST r32,imm32' which is the 16-bit
 * form of the instruction (addw %ax,=1234).  Many of the slots are
 * filled, but the VALID bit is not set.  In addition, there are some
 * instructions that will decode, but are simply not implemented by
 * the emulator (such as IDIV %ax,%cx).
 */
struct instructions instructions66[256] ={
  /* 00 */  { 0 },
  /* 01 */  { VALID | MODRM | WORD, { { add, "add" } } },
  /* 02 */  { 0 },
  /* 03 */  { VALID | MODRM | MR | WORD, { { add, "add" } } },
  /* 04 */  { 0 },
  /* 05 */  { IMM, { { add, "add" } } },
  /* 06 */  { 0 },
  /* 07 */  { 0 },
  /* 08 */  { 0 },
  /* 09 */  { VALID | MODRM | WORD, { { or, "or" } } },
  /* 0a */  { 0 },
  /* 0b */  { VALID | MODRM | MR | WORD, { { or, "or" } } },
  /* 0c */  { 0 },
  /* 0d */  { IMM, { { or, "or" } } },
  /* 0e */  { 0 },
  /* 0f */  { 0 },
  /* 10 */  { 0 },
  /* 11 */  { VALID | MODRM | WORD, { { adc, "adc" } } },
  /* 12 */  { 0 },
  /* 13 */  { VALID | MODRM | MR | WORD, { { adc, "adc" } } },
  /* 14 */  { 0 },
  /* 15 */  { IMM, { { adc, "adc" } } },
  /* 16 */  { 0 },
  /* 17 */  { 0 },
  /* 18 */  { 0 },
  /* 19 */  { VALID | MODRM | WORD, { { sbb, "sbb" } } },
  /* 1a */  { 0 },
  /* 1b */  { VALID | MODRM | MR | WORD, { { sbb, "sbb" } } },
  /* 1c */  { 0 },
  /* 1d */  { IMM, { { sbb, "sbb" } } },
  /* 1e */  { 0 },
  /* 1f */  { 0 },
  /* 20 */  { 0 },
  /* 21 */  { VALID | MODRM | WORD, { { and, "and" } } },
  /* 22 */  { 0 },
  /* 23 */  { VALID | MODRM | MR | WORD, { { and, "and" } } },
  /* 24 */  { 0 },
  /* 25 */  { IMM, { { and, "and" } } },
  /* 26 */  { 0 },
  /* 27 */  { 0 },
  /* 28 */  { 0 },
  /* 29 */  { VALID | MODRM | WORD, { { sub, "sub" } } },
  /* 2a */  { 0 },
  /* 2b */  { VALID | MODRM | MR | WORD, { { sub, "sub" } } },
  /* 2c */  { 0 },
  /* 2d */  { IMM, { { sub, "sub" } } },
  /* 2e */  { 0 },
  /* 2f */  { 0 },
  /* 30 */  { 0 },
  /* 31 */  { VALID | MODRM | WORD, { { xor, "xor" } } },
  /* 32 */  { 0 },
  /* 33 */  { VALID | MODRM | MR | WORD, { { xor, "xor" } } },
  /* 34 */  { 0 },
  /* 35 */  { IMM, { { xor, "xor" } } },
  /* 36 */  { 0 },
  /* 37 */  { 0 },
  /* 38 */  { 0 },
  /* 39 */  { VALID | MODRM | WORD, { { cmp, "cmp" } } },
  /* 3a */  { 0 },
  /* 3b */  { VALID | MODRM | MR | WORD, { { cmp, "cmp" } } },
  /* 3c */  { 0 },
  /* 3d */  { IMM, { { cmp, "cmp" } } },
  /* 3e */  { 0 },
  /* 3f */  { 0 },
  /* 40 */  { OPD, { { inc, "inc" } } },
  /* 41 */  { OPD, { { inc, "inc" } } },
  /* 42 */  { OPD, { { inc, "inc" } } },
  /* 43 */  { OPD, { { inc, "inc" } } },
  /* 44 */  { OPD, { { inc, "inc" } } },
  /* 45 */  { OPD, { { inc, "inc" } } },
  /* 46 */  { OPD, { { inc, "inc" } } },
  /* 47 */  { OPD, { { inc, "inc" } } },
  /* 48 */  { OPD, { { dec, "dec" } } },
  /* 49 */  { OPD, { { dec, "dec" } } },
  /* 4a */  { OPD, { { dec, "dec" } } },
  /* 4b */  { OPD, { { dec, "dec" } } },
  /* 4c */  { OPD, { { dec, "dec" } } },
  /* 4d */  { OPD, { { dec, "dec" } } },
  /* 4e */  { OPD, { { dec, "dec" } } },
  /* 4f */  { OPD, { { dec, "dec" } } },
  /* 50 */  { OPD, { { push, "push" } } },
  /* 51 */  { OPD, { { push, "push" } } },
  /* 52 */  { OPD, { { push, "push" } } },
  /* 53 */  { OPD, { { push, "push" } } },
  /* 54 */  { OPD, { { push, "push" } } },
  /* 55 */  { OPD, { { push, "push" } } },
  /* 56 */  { OPD, { { push, "push" } } },
  /* 57 */  { OPD, { { push, "push" } } },
  /* 58 */  { OPD, { { pop, "pop" } } },
  /* 59 */  { OPD, { { pop, "pop" } } },
  /* 5a */  { OPD, { { pop, "pop" } } },
  /* 5b */  { OPD, { { pop, "pop" } } },
  /* 5c */  { OPD, { { pop, "pop" } } },
  /* 5d */  { OPD, { { pop, "pop" } } },
  /* 5e */  { OPD, { { pop, "pop" } } },
  /* 5f */  { OPD, { { pop, "pop" } } },
  /* 60 */  { 0 },
  /* 61 */  { 0 },
  /* 62 */  { 0 },
  /* 63 */  { 0 },
  /* 64 */  { 0 },
  /* 65 */  { 0 },
  /* 66 */  { 0 },
  /* 67 */  { 0 },
  /* 68 */  { IMM, { { push, "push" } } },
  /* 69 */  { MR | IMM, { { imul, "imul" } } },
  /* 6a */  { 0 },
  /* 6b */  { 0 },
  /* 6c */  { 0 },
  /* 6d */  { 0 },
  /* 6e */  { 0 },
  /* 6f */  { 0 },
  /* 70 */  { 0 },
  /* 71 */  { 0 },
  /* 72 */  { 0 },
  /* 73 */  { 0 },
  /* 74 */  { 0 },
  /* 75 */  { 0 },
  /* 76 */  { 0 },
  /* 77 */  { 0 },
  /* 78 */  { 0 },
  /* 79 */  { 0 },
  /* 7a */  { 0 },
  /* 7b */  { 0 },
  /* 7c */  { 0 },
  /* 7d */  { 0 },
  /* 7e */  { 0 },
  /* 7f */  { 0 },
  /* 80 */  { 0 },
  /* 81 */  { VALID | GROUPR | IMM | WORD, { next: instructionsgroup1 } },
  /* 82 */  { 0 },
  /* 83 */  { VALID | GROUPR | IMM8 | WORD, { next: instructionsgroup1 } },
  /* 84 */  { 0 },
  /* 85 */  { VALID | MODRM | WORD, { { test, "test" } } },
  /* 86 */  { 0 },
  /* 87 */  { VALID | MODRM | WORD, { { xchg, "xchg" } } },
  /* 88 */  { 0 },
  /* 89 */  { VALID | MODRM | WORD, { { mov, "mov" } } },
  /* 8a */  { 0 },
  /* 8b */  { VALID | MODRM | MR | WORD, { { mov, "mov" } } },
  /* 8c */  { 0 },
  /* 8d */  { 0 },
  /* 8e */  { 0 },
  /* 8f */  { MODRM, { { pop, "pop" } } },
  /* 90 */  { 0 },
  /* 91 */  { OPD, { { xchg, "xchg" } } },
  /* 92 */  { OPD, { { xchg, "xchg" } } },
  /* 93 */  { OPD, { { xchg, "xchg" } } },
  /* 94 */  { OPD, { { xchg, "xchg" } } },
  /* 95 */  { OPD, { { xchg, "xchg" } } },
  /* 96 */  { OPD, { { xchg, "xchg" } } },
  /* 97 */  { OPD, { { xchg, "xchg" } } },
  /* 98 */  { VALID | DEFOP | WORD, { { cbw, "cbw" } } },
  /* 99 */  { DEFOP, { { cwd, "cwd" } } },
  /* 9a */  { 0 },
  /* 9b */  { 0 },
  /* 9c */  { 0, { { pushf, "pushf" } } },
  /* 9d */  { 0, { { popf, "popf" } } },
  /* 9e */  { 0 },
  /* 9f */  { 0 },
  /* a0 */  { 0 },
  /* a1 */  { 0 },
  /* a2 */  { 0 },
  /* a3 */  { 0 },
  /* a4 */  { 0 },
  /* a5 */  { 0 },
  /* a6 */  { 0 },
  /* a7 */  { 0 },
  /* a8 */  { 0 },
  /* a9 */  { IMM, { { test, "test" } } },
  /* aa */  { 0 },
  /* ab */  { 0 },
  /* ac */  { 0 },
  /* ad */  { 0 },
  /* ae */  { 0 },
  /* af */  { 0 },
  /* b0 */  { 0 },
  /* b1 */  { 0 },
  /* b2 */  { 0 },
  /* b3 */  { 0 },
  /* b4 */  { 0 },
  /* b5 */  { 0 },
  /* b6 */  { 0 },
  /* b7 */  { 0 },
  /* b8 */  { OPD | IMM, { { mov, "mov" } } },
  /* b9 */  { OPD | IMM, { { mov, "mov" } } },
  /* ba */  { OPD | IMM, { { mov, "mov" } } },
  /* bb */  { OPD | IMM, { { mov, "mov" } } },
  /* bc */  { OPD | IMM, { { mov, "mov" } } },
  /* bd */  { OPD | IMM, { { mov, "mov" } } },
  /* be */  { OPD | IMM, { { mov, "mov" } } },
  /* bf */  { OPD | IMM, { { mov, "mov" } } },
  /* c0 */  { 0 },
  /* c1 */  { VALID | IMM | GROUPR | WORD, { next: instructionsgroup2 } },
  /* c2 */  { 0 },
  /* c3 */  { 0 },
  /* c4 */  { 0 },
  /* c5 */  { 0 },
  /* c6 */  { 0 },
  /* c7 */  { VALID | MODRM | IMM | WORD, { { mov, "mov" } } },
  /* c8 */  { 0 },
  /* c9 */  { 0 },
  /* ca */  { 0 },
  /* cb */  { 0 },
  /* cc */  { 0 },
  /* cd */  { 0 },
  /* ce */  { 0 },
  /* cf */  { 0 },
  /* d0 */  { 0 },
  /* d1 */  { VALID | GROUPR | WORD, { next: instructionsgroup2 } },
  /* d2 */  { 0 },
  /* d3 */  { VALID | MODRM | GROUPR | WORD, { next: instructionsgroup2 } },
  /* d4 */  { 0 },
  /* d5 */  { 0 },
  /* d6 */  { 0 },
  /* d7 */  { 0, { { xlat, "xlat" } } },
  /* d8 */  { 0 },
  /* d9 */  { 0 },
  /* da */  { 0 },
  /* db */  { 0 },
  /* dc */  { 0 },
  /* dd */  { 0 },
  /* de */  { 0 },
  /* df */  { 0 },
  /* e0 */  { 0, { { loopne, "loopne" } } },
  /* e1 */  { 0, { { loope, "loope" } } },
  /* e2 */  { 0, { { loop, "loop" } } },
  /* e3 */  { REL8, { { jcxz, "jcxz" } } },
  /* e4 */  { 0 },
  /* e5 */  { 0 },
  /* e6 */  { 0 },
  /* e7 */  { 0 },
  /* e8 */  { REL32, { { call, "call" } } },
  /* e9 */  { REL32, { { jmp, "jmp" } } },
  /* ea */  { 0 },
  /* eb */  { REL8, { { jmp, "jmp" } } },
  /* ec */  { 0 },
  /* ed */  { 0 },
  /* ee */  { 0 },
  /* ef */  { 0 },
  /* f0 */  { 0 },
  /* f1 */  { 0 },
  /* f2 */  { 0 },
  /* f3 */  { 0 },
  /* f4 */  { 0 },
  /* f5 */  { 0 },
  /* f6 */  { 0 },
  /* f7 */  { VALID | GROUPR | WORD, { next: instructionsgroup3 } },
  /* f8 */  { 0 },
  /* f9 */  { 0 },
  /* fa */  { 0 },
  /* fb */  { 0 },
  /* fc */  { 0 },
  /* fd */  { 0 },
  /* fe */  { GROUPR, { next: instructionsgroup4 } },
  /* ff */  { VALID | GROUPR | WORD, { next: instructionsgroup5 } },
				       };

/*
 * Instruction opcode decode table
 *
 * All instructions not implemented by the emulator (for example, LES) are not
 * present in the table.
 */
struct instructions instructions[256] ={
  /* 00 */  { VALID | MODRM | BYTE, { { add, "add" } } },
  /* 01 */  { VALID | MODRM, { { add, "add" } } },
  /* 02 */  { VALID | MODRM | MR | BYTE, { { add, "add" } } },
  /* 03 */  { VALID | MODRM | MR, { { add, "add" } } },
  /* 04 */  { VALID | IMM | BYTE, { { add, "add" } } },
  /* 05 */  { VALID | IMM, { { add, "add" } } },
  /* 06 */  { 0 },
  /* 07 */  { 0 },
  /* 08 */  { VALID | MODRM | BYTE, { { or, "or" } } },
  /* 09 */  { VALID | MODRM, { { or, "or" } } },
  /* 0a */  { VALID | MODRM | MR | BYTE, { { or, "or" } } },
  /* 0b */  { VALID | MODRM | MR, { { or, "or" } } },
  /* 0c */  { VALID | IMM | BYTE, { { or, "or" } } },
  /* 0d */  { VALID | IMM, { { or, "or" } } },
  /* 0e */  { 0 },
  /* 0f */  { VALID | TWOBYTE, { next: instructions0F } },
  /* 10 */  { VALID | MODRM | BYTE, { { adc, "adc" } } },
  /* 11 */  { VALID | MODRM, { { adc, "adc" } } },
  /* 12 */  { VALID | MODRM | MR | BYTE, { { adc, "adc" } } },
  /* 13 */  { VALID | MODRM | MR, { { adc, "adc" } } },
  /* 14 */  { VALID | IMM | BYTE, { { adc, "adc" } } },
  /* 15 */  { VALID | IMM, { { adc, "adc" } } },
  /* 16 */  { 0 },
  /* 17 */  { 0 },
  /* 18 */  { VALID | MODRM | BYTE, { { sbb, "sbb" } } },
  /* 19 */  { VALID | MODRM, { { sbb, "sbb" } } },
  /* 1a */  { VALID | MODRM | MR | BYTE, { { sbb, "sbb" } } },
  /* 1b */  { VALID | MODRM | MR, { { sbb, "sbb" } } },
  /* 1c */  { VALID | IMM | BYTE, { { sbb, "sbb" } } },
  /* 1d */  { VALID | IMM, { { sbb, "sbb" } } },
  /* 1e */  { 0 },
  /* 1f */  { 0 },
  /* 20 */  { VALID | MODRM | BYTE, { { and, "and" } } },
  /* 21 */  { VALID | MODRM, { { and, "and" } } },
  /* 22 */  { VALID | MODRM | MR | BYTE, { { and, "and" } } },
  /* 23 */  { VALID | MODRM | MR, { { and, "and" } } },
  /* 24 */  { VALID | IMM | BYTE, { { and, "and" } } },
  /* 25 */  { VALID | IMM, { { and, "and" } } },
  /* 26 */  { 0 },
  /* 27 */  { VALID, { { daa, "daa" } } },
  /* 28 */  { VALID | MODRM | BYTE, { { sub, "sub" } } },
  /* 29 */  { VALID | MODRM, { { sub, "sub" } } },
  /* 2a */  { VALID | MODRM | MR | BYTE, { { sub, "sub" } } },
  /* 2b */  { VALID | MODRM | MR, { { sub, "sub" } } },
  /* 2c */  { VALID | IMM | BYTE, { { sub, "sub" } } },
  /* 2d */  { VALID | IMM, { { sub, "sub" } } },
  /* 2e */  { 0 },
  /* 2f */  { VALID, { { das, "das" } } },
  /* 30 */  { VALID | MODRM | BYTE, { { xor, "xor" } } },
  /* 31 */  { VALID | MODRM, { { xor, "xor" } } },
  /* 32 */  { VALID | MODRM | MR | BYTE, { { xor, "xor" } } },
  /* 33 */  { VALID | MODRM | MR, { { xor, "xor" } } },
  /* 34 */  { VALID | IMM | BYTE, { { xor, "xor" } } },
  /* 35 */  { VALID | IMM, { { xor, "xor" } } },
  /* 36 */  { 0 },
  /* 37 */  { VALID, { { aaa, "aaa" } } },
  /* 38 */  { VALID | MODRM | BYTE, { { cmp, "cmp" } } },
  /* 39 */  { VALID | MODRM, { { cmp, "cmp" } } },
  /* 3a */  { VALID | MODRM | MR | BYTE, { { cmp, "cmp" } } },
  /* 3b */  { VALID | MODRM | MR, { { cmp, "cmp" } } },
  /* 3c */  { VALID | IMM | BYTE, { { cmp, "cmp" } } },
  /* 3d */  { VALID | IMM, { { cmp, "cmp" } } },
  /* 3e */  { 0 },
  /* 3f */  { VALID, { { aas, "aas" } } },
  /* 40 */  { VALID | OPD, { { inc, "inc" } } },
  /* 41 */  { VALID | OPD, { { inc, "inc" } } },
  /* 42 */  { VALID | OPD, { { inc, "inc" } } },
  /* 43 */  { VALID | OPD, { { inc, "inc" } } },
  /* 44 */  { VALID | OPD, { { inc, "inc" } } },
  /* 45 */  { VALID | OPD, { { inc, "inc" } } },
  /* 46 */  { VALID | OPD, { { inc, "inc" } } },
  /* 47 */  { VALID | OPD, { { inc, "inc" } } },
  /* 48 */  { VALID | OPD, { { dec, "dec" } } },
  /* 49 */  { VALID | OPD, { { dec, "dec" } } },
  /* 4a */  { VALID | OPD, { { dec, "dec" } } },
  /* 4b */  { VALID | OPD, { { dec, "dec" } } },
  /* 4c */  { VALID | OPD, { { dec, "dec" } } },
  /* 4d */  { VALID | OPD, { { dec, "dec" } } },
  /* 4e */  { VALID | OPD, { { dec, "dec" } } },
  /* 4f */  { VALID | OPD, { { dec, "dec" } } },
  /* 50 */  { VALID | OPD, { { push, "push" } } },
  /* 51 */  { VALID | OPD, { { push, "push" } } },
  /* 52 */  { VALID | OPD, { { push, "push" } } },
  /* 53 */  { VALID | OPD, { { push, "push" } } },
  /* 54 */  { VALID | OPD, { { push, "push" } } },
  /* 55 */  { VALID | OPD, { { push, "push" } } },
  /* 56 */  { VALID | OPD, { { push, "push" } } },
  /* 57 */  { VALID | OPD, { { push, "push" } } },
  /* 58 */  { VALID | OPD, { { pop, "pop" } } },
  /* 59 */  { VALID | OPD, { { pop, "pop" } } },
  /* 5a */  { VALID | OPD, { { pop, "pop" } } },
  /* 5b */  { VALID | OPD, { { pop, "pop" } } },
  /* 5c */  { VALID | OPD, { { pop, "pop" } } },
  /* 5d */  { VALID | OPD, { { pop, "pop" } } },
  /* 5e */  { VALID | OPD, { { pop, "pop" } } },
  /* 5f */  { VALID | OPD, { { pop, "pop" } } },
  /* 60 */  { VALID, { { pusha, "pusha" } } },
  /* 61 */  { VALID, { { popa, "popa" } } },
  /* 62 */  { 0 },
  /* 63 */  { 0 },
  /* 64 */  { 0 },
  /* 65 */  { 0 },
  /* 66 */  { VALID | TWOBYTE, { next: instructions66 } },
  /* 67 */  { 0 },
  /* 68 */  { VALID | IMM, { { push, "push" } } },
  /* 69 */  { VALID | MR | IMM, { { imul, "imul" } } },
  /* 6a */  { VALID | BYTE | IMM, { { push, "push" } } },
  /* 6b */  { VALID | MR | BYTE | IMM, { { imul, "imul" } } },
  /* 6c */  { 0 },
  /* 6d */  { 0 },
  /* 6e */  { 0 },
  /* 6f */  { 0 },
  /* 70 */  { VALID | REL8, { { jcc, "jo" } } }, 
  /* 71 */  { VALID | REL8, { { jcc, "jno" } } },
  /* 72 */  { VALID | REL8, { { jcc, "jb" } } }, 
  /* 73 */  { VALID | REL8, { { jcc, "jae" } } },
  /* 74 */  { VALID | REL8, { { jcc, "jz" } } }, 
  /* 75 */  { VALID | REL8, { { jcc, "jnz" } } },
  /* 76 */  { VALID | REL8, { { jcc, "jbe" } } },
  /* 77 */  { VALID | REL8, { { jcc, "ja" } } }, 
  /* 78 */  { VALID | REL8, { { jcc, "js" } } }, 
  /* 79 */  { VALID | REL8, { { jcc, "jns" } } },
  /* 7a */  { VALID | REL8, { { jcc, "jpe" } } },
  /* 7b */  { VALID | REL8, { { jcc, "jpo" } } },
  /* 7c */  { VALID | REL8, { { jcc, "jl" } } }, 
  /* 7d */  { VALID | REL8, { { jcc, "jge" } } },
  /* 7e */  { VALID | REL8, { { jcc, "jle" } } },
  /* 7f */  { VALID | REL8, { { jcc, "jg" } } }, 
  /* 80 */  { VALID | GROUPR | IMM | BYTE, { next: instructionsgroup1 } },
  /* 81 */  { VALID | GROUPR | IMM, { next: instructionsgroup1 } },
  /* 82 */  { 0 },
  /* 83 */  { VALID | GROUPR | IMM8, { next: instructionsgroup1 } },
  /* 84 */  { VALID | MODRM | BYTE, { { test, "test" } } },
  /* 85 */  { VALID | MODRM, { { test, "test" } } },
  /* 86 */  { VALID | MODRM | BYTE, { { xchg, "xchg" } } },
  /* 87 */  { VALID | MODRM, { { xchg, "xchg" } } },
  /* 88 */  { VALID | MODRM | BYTE, { { mov, "mov" } } },
  /* 89 */  { VALID | MODRM, { { mov, "mov" } } },
  /* 8a */  { VALID | MODRM | MR | BYTE, { { mov, "mov" } } },
  /* 8b */  { VALID | MODRM | MR, { { mov, "mov" } } },
  /* 8c */  { 0 },
  /* 8d */  { VALID | MODRM | MR, { { lea, "lea" } } },
  /* 8e */  { 0 },
  /* 8f */  { VALID | MODRM, { { pop, "pop" } } },
  /* 90 */  { VALID, { { nop, "nop" } } },
  /* 91 */  { VALID | OPD, { { xchg, "xchg" } } },
  /* 92 */  { VALID | OPD, { { xchg, "xchg" } } },
  /* 93 */  { VALID | OPD, { { xchg, "xchg" } } },
  /* 94 */  { VALID | OPD, { { xchg, "xchg" } } },
  /* 95 */  { VALID | OPD, { { xchg, "xchg" } } },
  /* 96 */  { VALID | OPD, { { xchg, "xchg" } } },
  /* 97 */  { VALID | OPD, { { xchg, "xchg" } } },
  /* 98 */  { VALID | DEFOP, { { cbw, "cbw" } } },
  /* 99 */  { VALID | DEFOP, { { cwd, "cwd" } } },
  /* 9a */  { 0 },
  /* 9b */  { 0 },
  /* 9c */  { VALID, { { pushf, "pushf" } } },
  /* 9d */  { VALID, { { popf, "popf" } } },
  /* 9e */  { VALID | BYTE, { { sahf, "sahf" } } },
  /* 9f */  { VALID | BYTE, { { lahf, "lahf" } } },
  /* a0 */  { 0 },
  /* a1 */  { 0 },
  /* a2 */  { 0 },
  /* a3 */  { 0 },
  /* a4 */  { VALID | BYTE, { { movsb, "movsb" } } },
  /* a5 */  { VALID, { { movsd, "movsd" } } },
  /* a6 */  { VALID | BYTE, { { cmpsb, "cmpsb" } } },
  /* a7 */  { VALID, { { cmpsd, "cmpsd" } } },
  /* a8 */  { VALID | IMM | BYTE, { { test, "test" } } },
  /* a9 */  { VALID | IMM, { { test, "test" } } },
  /* aa */  { VALID | BYTE, { { stosb, "stosb" } } },
  /* ab */  { VALID, { { stosd, "stosd" } } },
  /* ac */  { VALID | BYTE, { { lodsb, "lodsb" } } },
  /* ad */  { VALID, { { lodsd, "lodsd" } } },
  /* ae */  { VALID | BYTE, { { scasb, "scasb" } } },
  /* af */  { VALID, { { scasd, "scasd" } } },
  /* b0 */  { VALID | OPD | IMM | BYTE, { { mov, "mov" } } },
  /* b1 */  { VALID | OPD | IMM | BYTE, { { mov, "mov" } } },
  /* b2 */  { VALID | OPD | IMM | BYTE, { { mov, "mov" } } },
  /* b3 */  { VALID | OPD | IMM | BYTE, { { mov, "mov" } } },
  /* b4 */  { VALID | OPD | IMM | BYTE, { { mov, "mov" } } },
  /* b5 */  { VALID | OPD | IMM | BYTE, { { mov, "mov" } } },
  /* b6 */  { VALID | OPD | IMM | BYTE, { { mov, "mov" } } },
  /* b7 */  { VALID | OPD | IMM | BYTE, { { mov, "mov" } } },
  /* b8 */  { VALID | OPD | IMM, { { mov, "mov" } } },
  /* b9 */  { VALID | OPD | IMM, { { mov, "mov" } } },
  /* ba */  { VALID | OPD | IMM, { { mov, "mov" } } },
  /* bb */  { VALID | OPD | IMM, { { mov, "mov" } } },
  /* bc */  { VALID | OPD | IMM, { { mov, "mov" } } },
  /* bd */  { VALID | OPD | IMM, { { mov, "mov" } } },
  /* be */  { VALID | OPD | IMM, { { mov, "mov" } } },
  /* bf */  { VALID | OPD | IMM, { { mov, "mov" } } },
  /* c0 */  { VALID | IMM | BYTE | GROUPR, { next: instructionsgroup2 } },
  /* c1 */  { VALID | IMM | GROUPR, { next: instructionsgroup2 } },
  /* c2 */  { VALID | IMM16, { { ret, "ret" } } },
  /* c3 */  { VALID, { { ret, "ret" } } },
  /* c4 */  { 0 },
  /* c5 */  { 0 },
  /* c6 */  { VALID | MODRM | IMM | BYTE, { { mov, "mov" } } },
  /* c7 */  { VALID | MODRM | IMM, { { mov, "mov" } } },
  /* c8 */  { 0 },
  /* c9 */  { 0 },
  /* ca */  { 0 },
  /* cb */  { 0 },
  /* cc */  { 0 },
  /* cd */  { 0 },
  /* ce */  { 0 },
  /* cf */  { 0 },
  /* d0 */  { VALID | BYTE | GROUPR, { next: instructionsgroup2 } },
  /* d1 */  { VALID | GROUPR, { next: instructionsgroup2 } },
  /* d2 */  { VALID | MODRM | BYTE | GROUPR, { next: instructionsgroup2 } },
  /* d3 */  { VALID | MODRM | GROUPR, { next: instructionsgroup2 } },
  /* d4 */  { VALID, { { aam, "aam" } } },
  /* d5 */  { VALID, { { aad, "aad" } } },
  /* d6 */  { 0 },
  /* d7 */  { VALID, { { xlat, "xlat" } } },
  /* d8 */  { 0 },
  /* d9 */  { 0 },
  /* da */  { 0 },
  /* db */  { 0 },
  /* dc */  { 0 },
  /* dd */  { 0 },
  /* de */  { 0 },
  /* df */  { 0 },
  /* e0 */  { VALID, { { loopne, "loopne" } } },
  /* e1 */  { VALID, { { loope, "loope" } } },
  /* e2 */  { VALID, { { loop, "loop" } } },
  /* e3 */  { VALID | REL8, { { jcxz, "jcxz" } } },
  /* e4 */  { 0 },
  /* e5 */  { 0 },
  /* e6 */  { 0 },
  /* e7 */  { 0 },
  /* e8 */  { VALID | REL32, { { call, "call" } } },
  /* e9 */  { VALID | REL32, { { jmp, "jmp" } } },
  /* ea */  { 0 },
  /* eb */  { VALID | REL8, { { jmp, "jmp" } } },
  /* ec */  { 0 },
  /* ed */  { 0 },
  /* ee */  { 0 },
  /* ef */  { 0 },
  /* f0 */  { 0 },
  /* f1 */  { 0 },
  /* f2 */  { VALID, { { repne, "repne" } } },
  /* f3 */  { VALID, { { rep, "rep" } } },
  /* f4 */  { 0 },
  /* f5 */  { VALID, { { cmc, "cmc" } } },
  /* f6 */  { VALID | BYTE | GROUPR, { next: instructionsgroup3 } },
  /* f7 */  { VALID | GROUPR, { next: instructionsgroup3 } },
  /* f8 */  { VALID, { { clc, "clc" } } },
  /* f9 */  { VALID, { { stc, "stc" } } },
  /* fa */  { 0 },
  /* fb */  { 0 },
  /* fc */  { VALID, { { cld, "cld" } } },
  /* fd */  { VALID, { { std, "std" } } },
  /* fe */  { VALID | GROUPR, { next: instructionsgroup4 } },
  /* ff */  { VALID | GROUPR, { next: instructionsgroup5 } },
				       };

/*
 * Function prototypes
 */
void		abend( int, struct sigframe* );
void		fault( struct iAPX486*, int );
unsigned char	*modrm( unsigned, struct modrm*, unsigned**, unsigned*, char*,
		        unsigned**, unsigned*, char*, struct iAPX486* );
unsigned char	*sib( unsigned, unsigned, struct sib*, unsigned**,
		      unsigned*, char*, struct iAPX486* );
char		*theflags( unsigned );


/*
 * Fetch/store an operand, taking into account its size
 */
#define	fetch(flag,addr)	(((flag)&BYTE)?           (*(unsigned char*)(addr)) \
					: (((flag)&WORD)? (*(unsigned short*)(addr)) \
							: (*(unsigned int*)(addr))))

#define	fetchs(flag,addr)	(((flag)&BYTE)?           (*(signed char*)(addr)) \
					: (((flag)&WORD)? (*(short*)(addr)) \
							: (*(int*)(addr))))

#define	store(flag,addr,data)	(((flag)&BYTE)?           (*(unsigned char*)(addr) = (data)) \
					: (((flag)&WORD)? (*(unsigned short*)(addr) = (data)) \
							: (*(unsigned int*)(addr) = (data))))


/*
 * Allow space to print the instruction
 */
#define	IWIDTH	22

/*
 * The emulator
 *
 * Emulate a single instruction.  STATE contains all register and flag values.
 * No address validity checking is done before fetching/storing a value.
 */
 int
iAPX486( register struct iAPX486 *state )
{
  struct instructions *ip;
  unsigned char *p, b1, b2;
  unsigned flags, size, *src, srcv, *dest, destv;
  int r, n, x, v1, v2;
  char *reg, ops[32], op1[32], op2[32];

  /*
   * Remember any instruction prefixes
   */
  static unsigned prefix = 0;

  /*
   * Decode and emulate the next instruction
   */
  p = (unsigned char*) state->eip;

  ip = &instructions[ *p++ ];

  if ( ! (ip->flags & VALID) )
    {
      fault( state, SIGIOT );
      return -1;
    }
  
  if ( ip->flags & TWOBYTE )
    {
      ip = &ip->u.next[ *p++ ];

      if ( ! (ip->flags & VALID) )
	{
	  fault( state, SIGIOT );
	  return -1;
	}
    }

  flags = ip->flags;
  
  if ( ip->flags & GROUPR )
    {
      ip = &ip->u.next[ ((struct modrm*) p)->reg ];

      if ( ! (ip->flags & VALID) )
	{
	  fault( state, SIGIOT );
	  return -1;
	}

      flags |= ip->flags;
    }

  r = eax;
  if ( flags & OPD )
    r = p[-1] & 0x07;

  if ( flags & BYTE )
    {
      size = 2;
      reg = regname8[r];
    }
  else
    if ( flags & WORD )
      {
	size = 4;
	reg = regname16[r];
      }
    else
      {
	size = 8;
	reg = regname32[r];
      }

  if ( ! (flags & DEFOP) )
    {
      switch( ip->u.instruction.code )
        {
      case cmpsb:
      case cmpsd:
      case lodsb:
      case lodsd:
      case movsb:
      case movsd:
      case scasb:
      case scasd:
      case stosb:
      case stosd:
	  n = printf( "%8p:  %-8s", (unsigned*) state->eip, ip->u.instruction.opcode );
          break;
      default:
	  sprintf( ops, "%s%.1s", ip->u.instruction.opcode, "..b.w...\0"+size );
	  n = printf( "%8p:  %-8s", (unsigned*) state->eip, ops );
          break;
	}
    }

  switch( ip->u.instruction.code )
    {
  case adc:
      if ( flags & IMM8 )
        /* 83 */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  x = *(signed char*)p;
	  sprintf( ops, "%s,%d", op1, x );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; adcl %3,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv), "g" (x)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %08x\n", n+IWIDTH, "", op1, destv );
	  printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  state->eip = (unsigned) p + 1;
	}
      else
	if ( flags & IMM )
	  /* 1[45], 8[01] */
	  {
	    if ( flags & MODRM )
	      /* 8[01] */
	      {
		p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

		x = fetchs( flags, p );
		sprintf( ops, "%s,%d", op1, x );
		printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

		destv = fetchs( flags, dest );
		asm( "movl %2,%%eax; adcl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, dest, destv );

		printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip = (unsigned) p + (size >> 1);
	      }
	    else
	      /* 1[45] */
	      {
		x = fetchs( flags, p );
		sprintf( ops, "%%%s,%d", reg, x );
		printf( "%-*s    %%eax = %08x\n", IWIDTH, ops, state->registers[eax] );

		destv = fetchs( flags, &state->registers[eax] );
		asm( "movl %2,%%eax; adcl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, &state->registers[eax], destv );

		printf( "%*s ## %%eax = %08x\n", n+IWIDTH, "", state->registers[eax] );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip += 1 + (size >> 1);
	      }
	  }
	else
	  /* 1[0123] */
	  {
	    p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	    sprintf( ops, "%s,%s", op1, op2 );
	    printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );
	    printf( "%*s    %s = %0*x\n", n+IWIDTH, "", op2, size, srcv );

	    destv = fetchs( flags, dest );
	    srcv = fetchs( flags, src );
	    asm( "movl %2,%%eax; adcl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		 : "=g" (destv), "=m" (state->efl)
		 : "0" (destv), "g" (srcv)
		 : "%eax", "cc"
	       );
	    store( flags, dest, destv );

	    printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	    printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	    state->eip = (unsigned) p;
	  }

      break;
  case add:
      if ( flags & IMM8 )
        /* 83 */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  x = *(signed char*)p;
	  sprintf( ops, "%s,%d", op1, x );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; addl %3,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv), "g" (x)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %08x\n", n+IWIDTH, "", op1, destv );
	  printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  state->eip = (unsigned) p + 1;
	}
      else
	if ( flags & IMM )
	  /* 0[45], 8[01] */
	  {
	    if ( flags & MODRM )
	      /* 8[01] */
	      {
		p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

		x = fetchs( flags, p );
		sprintf( ops, "%s,%d", op1, x );

		destv = fetchs( flags, dest );
		printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

		asm( "movl %2,%%eax; addl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, dest, destv );

		printf( "%*s ## %s = %08x\n", n+IWIDTH, "", op1, destv );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip = (unsigned) p + 4;
	      }
	    else
	      /* 0[45] */
	      {
		x = fetchs( flags, p );
		sprintf( ops, "%%%s,%d", reg, x );
		printf( "%-*s    %%eax = %08x\n", IWIDTH, ops, state->registers[eax] );

		destv = fetchs( flags, &state->registers[eax] );
		asm( "movl %2,%%eax; addl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, &state->registers[eax], destv );

		printf( "%*s ## %%eax = %08x\n", n+IWIDTH, "", state->registers[eax] );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip += 1 + (size >> 1);
	      }
	  }
	else
	  /* 0[0123] */
	  {
	    p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	    sprintf( ops, "%s,%s", op1, op2 );
	    printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );
	    printf( "%*s    %s = %08x\n", n+IWIDTH, "", op2, srcv );

	    destv = fetchs( flags, dest );
	    srcv = fetchs( flags, src );
	    asm( "movl %2,%%eax; addl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		 : "=g" (destv), "=m" (state->efl)
		 : "0" (destv), "g" (srcv)
		 : "%eax", "cc"
	       );
	    store( flags, dest, destv );

	    printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	    printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	    state->eip = (unsigned) p;
	  }

      break;
  case and:
      if ( flags & IMM8 )
        /* 83 */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  x = *(signed char*)p;
	  sprintf( ops, "%s,%d", op1, x );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; andl %3,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv), "g" (x)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %08x\n", n+IWIDTH, "", op1, destv );
	  printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  state->eip = (unsigned) p + 1;
	}
      else
	if ( flags & IMM )
	  /* 2[45], 8[01] */
	  {
	    if ( flags & MODRM )
	      /* 8[01] */
	      {
		p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

		x = fetchs( flags, p );
		sprintf( ops, "%s,%d", op1, x );

		destv = fetchs( flags, dest );
		printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

		asm( "movl %2,%%eax; andl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, dest, destv );

		printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip = (unsigned) p + (size >> 1);
	      }
	    else
	      /* 2[45] */
	      {
		x = fetchs( flags, p );
		sprintf( ops, "%%%s,%d", reg, x );
		printf( "%-*s    %%eax = %08x\n", IWIDTH, ops, state->registers[eax] );

		destv = fetchs( flags, &state->registers[eax] );
		asm( "movl %2,%%eax; andl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, &state->registers[eax], destv );

		printf( "%*s ## %%eax = %08x\n", n+IWIDTH, "", state->registers[eax] );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip += 1 + (size >> 1);
	      }
	  }
	else
	  /* 2[0123] */
	  {
	    p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	    sprintf( ops, "%s,%s", op1, op2 );
	    printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );
	    printf( "%*s    %s = %0*x\n", n+IWIDTH, "", op2, size, srcv );

	    destv = fetchs( flags, dest );
	    srcv = fetchs( flags, src );
	    asm( "movl %2,%%eax; andl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		 : "=g" (destv), "=m" (state->efl)
		 : "0" (destv), "g" (srcv)
		 : "%eax", "cc"
	       );
	    store( flags, dest, destv );

	    printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	    printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	    state->eip = (unsigned) p;
	  }

      break;
  case bswap:
      printf( "%%%-*s    %%%s = %08x\n", IWIDTH-1, regname32[r], regname32[r], state->registers[r] );
      x = state->registers[r];
      x = (x << 24) | ((x << 8) & 0x00ff0000) | ((x >> 8) & 0x0000ff00) | ((x >> 24) & 0x000000ff);
      state->registers[r] = x;
      printf( "%*s ## %%%s = %08x\n", n+IWIDTH, "", regname32[r], state->registers[r] );
      state->eip += 2;
      break;
  case bt:
      if ( flags & IMM8 )
	/* 0fba */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  x = *(unsigned char*)p & 0x1f;
	  sprintf( ops, "%s,%d", op1, x );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );
	  b2 = x;
	  asm( "movl %1,%%eax; bt %%eax,%2; pushf; pop %0"
	       : "=m" (state->efl)
	       : "g" (x), "g" (destv)
	       : "%eax", "cc"
	     );
	  state->eip = (unsigned) p + 1;
	}
      else
	/* 0fab */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	  sprintf( ops, "%s,%s", op1, op2 );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );
	  printf( "%*s    %s = %08x\n", n+IWIDTH, "", op2, srcv );

	  asm( "movl %1,%%eax; bt %%eax,%2; pushf; pop %0"
	       : "=m" (state->efl)
	       : "g" (srcv), "g" (destv)
	       : "%eax", "cc"
	     );
	  state->eip = (unsigned) p;
	}

      *dest = destv;
      printf( "%*s ## %s = %08x\n", n+IWIDTH, "", op1, destv );
      printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );

      break;
  case btc:
      if ( flags & IMM8 )
	/* 0fba */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  x = *(unsigned char*)p & 0x1f;
	  sprintf( ops, "%s,%d", op1, x );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );
	  b2 = x;
	  asm( "movl %1,%%eax; btc %%eax,%2; pushf; pop %0"
	       : "=m" (state->efl)
	       : "g" (x), "g" (destv)
	       : "%eax", "cc"
	     );
	  state->eip = (unsigned) p + 1;
	}
      else
	/* 0fab */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	  sprintf( ops, "%s,%s", op1, op2 );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );
	  printf( "%*s    %s = %08x\n", n+IWIDTH, "", op2, srcv );

	  asm( "movl %1,%%eax; btc %%eax,%2; pushf; pop %0"
	       : "=m" (state->efl)
	       : "g" (srcv), "g" (destv)
	       : "%eax", "cc"
	     );
	  state->eip = (unsigned) p;
	}

      *dest = destv;
      printf( "%*s ## %s = %08x\n", n+IWIDTH, "", op1, destv );
      printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );

      break;
  case btr:
      if ( flags & IMM8 )
	/* 0fba */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  x = *(unsigned char*)p & 0x1f;
	  sprintf( ops, "%s,%d", op1, x );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );
	  b2 = x;
	  asm( "movl %1,%%eax; btr %%eax,%2; pushf; pop %0"
	       : "=m" (state->efl)
	       : "g" (x), "g" (destv)
	       : "%eax", "cc"
	     );
	  state->eip = (unsigned) p + 1;
	}
      else
	/* 0fab */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	  sprintf( ops, "%s,%s", op1, op2 );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );
	  printf( "%*s    %s = %08x\n", n+IWIDTH, "", op2, srcv );

	  asm( "movl %1,%%eax; btr %%eax,%2; pushf; pop %0"
	       : "=m" (state->efl)
	       : "g" (srcv), "g" (destv)
	       : "%eax", "cc"
	     );
	  state->eip = (unsigned) p;
	}

      *dest = destv;
      printf( "%*s ## %s = %08x\n", n+IWIDTH, "", op1, destv );
      printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );

      break;
  case bts:
      if ( flags & IMM8 )
	/* 0fba */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  x = *(unsigned char*)p & 0x1f;
	  sprintf( ops, "%s,%d", op1, x );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );
	  b2 = x;
	  asm( "movl %1,%%eax; bts %%eax,%2; pushf; pop %0"
	       : "=m" (state->efl)
	       : "g" (x), "g" (destv)
	       : "%eax", "cc"
	     );
	  state->eip = (unsigned) p + 1;
	}
      else
	/* 0fab */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	  sprintf( ops, "%s,%s", op1, op2 );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );
	  printf( "%*s    %s = %08x\n", n+IWIDTH, "", op2, srcv );

	  asm( "movl %1,%%eax; bts %%eax,%2; pushf; pop %0"
	       : "=m" (state->efl)
	       : "g" (srcv), "g" (destv)
	       : "%eax", "cc"
	     );
	  state->eip = (unsigned) p;
	}

      *dest = destv;
      printf( "%*s ## %s = %08x\n", n+IWIDTH, "", op1, destv );
      printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );

      break;
  case call:
      if ( flags & REL32 )
	{
	  x = *(int*)p;
	  printf( "%-*p    %%esp = %8p\n", IWIDTH, (unsigned*) (state->eip = (unsigned)p + 4 + x),
						    (unsigned*) state->registers[esp] );
	  * (unsigned*) (state->registers[esp] -= 4) = (unsigned) p + 4;
	}
      else
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );
	  printf( "%-*p    %%esp = %8p\n", IWIDTH, (unsigned*) (state->eip = destv),
							(unsigned*) state->registers[esp] );
	  * (unsigned*) (state->registers[esp] -= 4) = (unsigned) p;
	}
      printf( "%*s ## %%esp = %8p  ->  %08x\n", n+IWIDTH, "",
      		(unsigned*) (state->registers[esp]), ((unsigned*) state->registers[esp])[0] );
      break;
  case cbw:
      printf( "%8p:  cwde    %n%*s    %%ax = %04x\n",
      		(unsigned*) state->eip, &n, IWIDTH, "", * (short*) &state->registers[eax] );
      state->registers[eax] = * (short*) &state->registers[eax];
      printf( "%*s ## %%eax = %08x\n", n+IWIDTH, "", state->registers[eax] );
      state->eip += 1;
      break;
  case clc:
      state->efl &= ~0x00000001;
      printf( "%*s ## efl: %s\n", IWIDTH, "", theflags(state->efl) );
      state->eip += 1;
      break;
  case cld:
      state->efl &= ~0x00000400;
      printf( "%*s ## efl: %s\n", IWIDTH, "", theflags(state->efl) );
      state->eip += 1;
      break;
  case cmc:
      state->efl ^= 0x00000001;
      printf( "%*s ## efl: %s\n", IWIDTH, "", theflags(state->efl) );
      state->eip += 1;
      break;
  case cmp:
      if ( flags & IMM8 )
        /* 83 */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  x = *(signed char*)p;
	  sprintf( ops, "%s,%d", op1, x );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );

	  destv = fetchs( flags, dest );
	  asm( "movl %1,%%eax; cmpl %2,%%eax; pushf; pop %0"
	       : "=m" (state->efl)
	       : "g" (destv), "g" (x)
	       : "%eax", "cc"
	     );

	  printf( "%*s ## efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  state->eip = (unsigned) p + 1;
	}
      else
	if ( flags & IMM )
	  /* 3[cd], 8[01] */
	  {
	    if ( flags & MODRM )
	      /* 8[01] */
	      {
		p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

		x = fetchs( flags, p );
		sprintf( ops, "%s,%d", op1, x );
		printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

		destv = fetchs( flags, dest );
		asm( "movl %1,%%eax; cmpl %2,%%eax; pushf; pop %0"
		     : "=m" (state->efl)
		     : "g" (destv), "g" (x)
		     : "%eax", "cc"
		   );

		printf( "%*s ## efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip = (unsigned) p + (size >> 1);
	      }
	    else
	      /* 3[cd] */
	      {
		x = fetchs( flags, p );
		sprintf( ops, "%%%s,%d", reg, x );
		printf( "%-*s    %%eax = %08x\n", IWIDTH, ops, state->registers[eax] );

		destv = fetchs( flags, dest );
		asm( "movl %1,%%eax; cmpl %2,%%eax; pushf; pop %0"
		     : "=m" (state->efl)
		     : "g" (destv), "g" (x)
		     : "%eax", "cc"
		   );

		printf( "%*s ## efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip += 1 + (size >> 1);
	      }
	  }
	else
	  /* 3[89ab] */
	  {
	    p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	    sprintf( ops, "%s,%s", op1, op2 );
	    printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );
	    printf( "%*s    %s = %0*x\n", n+IWIDTH, "", op2, size, srcv );

	    destv = fetchs( flags, dest );
	    srcv = fetchs( flags, src );
	    asm( "movl %1,%%eax; cmpl %2,%%eax; pushf; pop %0"
		 : "=m" (state->efl)
		 : "g" (destv), "g" (srcv)
		 : "%eax", "cc"
	       );

	    printf( "%*s ## efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	    state->eip = (unsigned) p;
	  }

      break;
  case cmpsb:
      if ( prefix & (REP | REPNE) )
        {
	  if ( state->registers[ecx] != 0 )
	    {
	      for( r = 0; state->registers[ecx] != 0; r = 1 )
	        {
		  if ( r )
		    printf( "%*s    %%ecx = %08x\n", n+IWIDTH, "", state->registers[ecx] );
		  else
		    printf( "%*s    %%ecx = %08x\n", IWIDTH, "", state->registers[ecx] );

		  printf( "%*s    [%%esi] = %02x\n", n+IWIDTH, "",
					      * (unsigned char*) state->registers[esi] );
		  printf( "%*s    [%%edi] = %02x\n", n+IWIDTH, "",
					      * (unsigned char*) state->registers[edi] );

		  destv = fetchs( flags, state->registers[edi] );
		  srcv = fetchs( flags, state->registers[esi] );
		  asm( "movl %1,%%eax; cmpl %%eax,%2; pushf; pop %0"
		       : "=m" (state->efl)
		       : "g" (destv), "g" (srcv)
		       : "%eax", "cc"
		     );

		  printf( "%*s  # efl: %s\n", n+IWIDTH, "", theflags(state->efl) );

		  if ( state->efl & 0x00000400 )
		    {
		      state->registers[edi]--;
		      state->registers[esi]--;
		    }
		  else
		    {
		      state->registers[edi]++;
		      state->registers[esi]++;
		    }

		  state->registers[ecx]--;

		  if ( prefix & REP && ! (state->efl & 0x00000040) )
		    break;
		  if ( prefix & REPNE && state->efl & 0x00000040 )
		    break;
		}

	      printf( "%*s ## %%esi = %08x\n", n+IWIDTH, "", state->registers[esi] );
	      printf( "%*s    %%edi = %08x\n", n+IWIDTH, "", state->registers[edi] );
	    }
	  else
	    printf( "%*s    %%ecx = 0\n", IWIDTH, "" );
	}
      else
	{
	  printf( "%*s    [%%esi] = %02x\n", IWIDTH, "", * (unsigned char*) state->registers[esi] );
	  printf( "%*s    [%%edi] = %02x\n", n+IWIDTH, "", * (unsigned char*) state->registers[edi] );

	  destv = fetchs( flags, state->registers[edi] );
	  srcv = fetchs( flags, state->registers[esi] );
	  asm( "movl %1,%%eax; cmpl %%eax,%2; pushf; pop %0"
	       : "=m" (state->efl)
	       : "g" (destv), "g" (srcv)
	       : "%eax", "cc"
	     );

	  if ( state->efl & 0x00000400 )
	    {
	      state->registers[edi]--;
	      state->registers[esi]--;
	    }
	  else
	    {
	      state->registers[edi]++;
	      state->registers[esi]++;
	    }

	  printf( "%*s ## efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  printf( "%*s    %%esi = %08x\n", n+IWIDTH, "", state->registers[esi] );
	  printf( "%*s    %%edi = %08x\n", n+IWIDTH, "", state->registers[edi] );
	}

      state->eip += 1;
      break;
  case cwd:
      printf( "%8p:  cdq     %n%*s    %%eax = %08x\n",
				  (unsigned*) state->eip, &n, IWIDTH, "", state->registers[eax] );
      if ( state->registers[eax] & 0x80000000 )
        state->registers[edx] = -1;
      else
        state->registers[edx] = 0;
      printf( "%*s ## %%edx = %08x\n", n+IWIDTH, "", state->registers[edx] );
      printf( "%*s    %%eax = %08x\n", n+IWIDTH, "", state->registers[eax] );
      state->eip += 1;
      break;
  case dec:
      if ( flags & MODRM )
        /* f[ef] */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  destv = fetchs( flags, dest );
	  printf( "%-*s    %s = %0*x\n", IWIDTH, op1, op1, size, destv );
	  asm( "movl %2,%%eax; decl %%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	}
      else
        /* 40 */
        {
	  destv = state->registers[r];
	  printf( "%%%-*s    %%%s = %08x\n", IWIDTH-1, regname32[r], regname32[r], state->registers[r] );

	  asm( "movl %2,%%eax; decl %%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv)
	       : "%eax", "cc"
	     );
	  state->registers[r] = destv;

	  printf( "%*s ## %%%s = %08x\n", n+IWIDTH, "", regname32[r], state->registers[r] );
	}

      printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
      state->eip = (unsigned) p;
      break;
  case idiv:
      p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );
      printf( "%-*s    %%edx:%%eax = %08x %08x\n", IWIDTH, op1,
      					v1 = state->registers[edx], v2 = state->registers[eax] );
      printf( "%*s    %s = %08x\n", n+IWIDTH, "", op1, destv );
      asm( "movl %3,%%edx; movl %4,%%eax; movl %2,%%ecx; idiv %%ecx; movl %%edx,%0; movl %%eax,%1"
	   : "=g" (destv), "=g" (srcv)
	   : "0" (destv), "g" (v1), "g" (v2)
	   : "%eax", "edx", "ecx", "cc"
	 );
      printf( "%*s ## %%edx = %08x\n", n+IWIDTH, "", state->registers[edx] = destv );
      printf( "%*s    %%eax = %08x\n", n+IWIDTH, "", state->registers[eax] = srcv );
      state->eip = (unsigned) p;
      break;
  case imul:
      p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

      if ( flags & MODRM )
        /* 0faf */
	{
	  sprintf( ops, "%s,%s", op1, op2 );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );
	  printf( "%*s    %s = %08x\n", n+IWIDTH, "", op2, srcv );
	  asm( "movl %2,%%eax; movl %3,%%ecx; imul %%ecx,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv), "g" (srcv)
	       : "%ecx", "%eax", "cc"
	     );
	  *dest = destv;
	  printf( "%*s ## %s = %08x\n", n+IWIDTH, "", op1, destv );
	  state->eip = (unsigned) p;
	}
      else
        if ( flags & IMM )
	  /* 6[9b] */
	  {
	    if ( flags & BYTE )
	      {
		x = *(signed char*)p;
		state->eip = (unsigned) p + 1;
	      }
	    else
	      {
		x = *(int*)p;
		state->eip = (unsigned) p + 4;
	      }
	    sprintf( ops, "%s,%s,%d", op1, op2, x );
	    printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );
	    printf( "%*s    %s = %08x\n", n+IWIDTH, "", op2, srcv );
	    asm( "movl %2,%%eax; movl %3,%%ecx; imul %%ecx,%%eax; movl %%eax,%0; pushf; pop %1"
		 : "=g" (destv), "=m" (state->efl)
		 : "g" (srcv), "g" (x)
		 : "%eax", "ecx", "cc"
	       );
	  *dest = destv;
	  printf( "%*s ## %s = %08x\n", n+IWIDTH, "", op1, destv );
	  }
	else
	  /* f7 */
	  {
	    printf( "%-*s    %%eax = %08x\n", IWIDTH, op1, x = state->registers[eax] );
	    printf( "%*s    %s = %08x\n", n+IWIDTH, "", op1, destv );
	    asm( "movl %4,%%eax; movl %3,%%ecx; imul %%ecx; movl %%edx,%0; movl %%eax,%1; pushf; pop %2"
		 : "=g" (v1), "=g" (v2), "=m" (state->efl)
		 : "g" (destv), "g" (x)
		 : "%ecx", "%edx", "%eax", "cc"
	       );
	    printf( "%*s ## %%edx:%%eax = %08x %08x\n", n+IWIDTH, "",
	    					state->registers[edx] = v1,
						state->registers[eax] = v2 );
	    state->eip = (unsigned) p;
	  }
      printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
      break;
  case inc:
      if ( flags & MODRM )
        /* f[ef] */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  printf( "%-*s    %s = %0*x\n", IWIDTH, op1, op1, size, destv );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; incl %%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	}
      else
        /* 40 */
        {
	  printf( "%%%-*s    %%%s = %08x\n", IWIDTH-1, regname32[r], regname32[r], state->registers[r] );

	  destv = state->registers[r];
	  asm( "movl %2,%%eax; incl %%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv)
	       : "%eax", "cc"
	     );
	  state->registers[r] = destv;

	  printf( "%*s ## %%%s = %08x\n", n+IWIDTH, "", regname32[r], state->registers[r] );
	}

      printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
      state->eip = (unsigned) p;
      break;
  case jcc:
      r = p[-1] & 0x0f;

      if ( flags & REL8 )
        {
	  x = *(signed char*)p;
	  printf( "%-*p", IWIDTH, (unsigned*) (x = (unsigned)p + 1 + x) );
	  ++p;
	}
      else
	{
	  x = *(int*)p;
	  printf( "%-*p", IWIDTH, (unsigned*) (x = (unsigned)p + 4 + x) );
	  p += 4;
	}
      printf( "    efl: %s\n", theflags(state->efl) );
      switch( r )
        {
      case 0:
	  asm( "sub %%eax,%%eax; push %1; popf; jo 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 1:
	  asm( "sub %%eax,%%eax; push %1; popf; jno 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 2:
	  asm( "sub %%eax,%%eax; push %1; popf; jb 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 3:
	  asm( "sub %%eax,%%eax; push %1; popf; jae 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 4:
	  asm( "sub %%eax,%%eax; push %1; popf; jz 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 5:
	  asm( "sub %%eax,%%eax; push %1; popf; jnz 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 6:
	  asm( "sub %%eax,%%eax; push %1; popf; jbe 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 7:
	  asm( "sub %%eax,%%eax; push %1; popf; ja 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 8:
	  asm( "sub %%eax,%%eax; push %1; popf; js 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 9:
	  asm( "sub %%eax,%%eax; push %1; popf; jns 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 10:
	  asm( "sub %%eax,%%eax; push %1; popf; jpe 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 11:
	  asm( "sub %%eax,%%eax; push %1; popf; jpo 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 12:
	  asm( "sub %%eax,%%eax; push %1; popf; jl 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 13:
	  asm( "sub %%eax,%%eax; push %1; popf; jge 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 14:
	  asm( "sub %%eax,%%eax; push %1; popf; jle 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 15:
	  asm( "sub %%eax,%%eax; push %1; popf; jg 0f; inc %%eax; 0: movl %%eax,%0"
	       : "=g" (r)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
	}
      if ( r == 0 )
	printf( "%*s ## eip = %8p\n", n+IWIDTH, "", (unsigned*) (state->eip = x) );
      else
        state->eip = (unsigned) p;
      break;
  case jcxz:
      x = *(signed char*)p;
      printf( "%-*p", IWIDTH, (unsigned*) (x = (unsigned)p + 1 + x) );
      ++p;
      printf( "    %%ecx = %08x\n", state->registers[ecx] );
      if ( state->registers[ecx] == 0 )
	printf( "%*s ## eip = %8p\n", n+IWIDTH, "", (unsigned*) (state->eip = x) );
      else
        state->eip += 2;
      break;
  case jmp:
      if ( flags & REL8 )
        {
	  x = *(signed char*)p;
	  printf( "%-8p\n", (unsigned*) (state->eip = (unsigned)p + 1 + x) );
	}
      else
	if ( flags & REL32 )
	  {
	    x = *(int*)p;
	    printf( "%-8p\n", (unsigned*) (state->eip = (unsigned)p + 4 + x) );
	  }
	else
	  {
	    p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );
	    printf( "%-*s ## eip = %-8p\n", IWIDTH, op1, (unsigned*) (state->eip = destv) );
	  }
      break;
  case lea:
      p = modrm( NOV | flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

      sprintf( ops, "%s,%s", op1, op2 );
      printf( "%-*s ## %s = %08x\n", IWIDTH, ops, op1, *dest = (unsigned) src );

      state->eip = (unsigned) p;
      break;
  case mov:
      if ( flags & OPD )
	{
	  if ( flags & BYTE )
	    {
	      dest = (unsigned*) ((char*) &state->registers[r&0x03] + ((r & 0x04)? 1 : 0));
	      r &= 0x03;
	    }
	  else
	    dest = &state->registers[r];

	  x = fetchs( flags, p );
	  sprintf( ops, "%%%s,%d", reg, x );
	  printf( "%-*s    %%%s = %08x\n", IWIDTH, ops, regname32[r], state->registers[r] );

	  store( flags, dest, x );

	  printf( "%*s ## %%%s = %08x\n", n+IWIDTH, "", regname32[r], state->registers[r] );
	  state->eip += 1 + (size >> 1);
	}
      else
        {
	  if ( flags & IMM )
	    {
	      p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	      x = fetchs( flags, p );
	      sprintf( ops, "%s,%d", op1, x );
	      printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

	      store( flags, dest, x );

	      printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, x );
	      state->eip = (unsigned) p + (size >> 1);
	    }
	  else
	    {
	      p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	      sprintf( ops, "%s,%s", op1, op2 );
	      printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );
	      printf( "%*s    %s = %0*x\n", n+IWIDTH, "",  op2, size, srcv );

	      store( flags, dest, srcv );

	      printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, srcv );
	      state->eip = (unsigned) p;
	    }
	}

      break;
  case movsb:
      if ( prefix & REP )
        {
	  printf( "%*s    %%ecx = %08x\n", IWIDTH, "", state->registers[ecx] );

	  while( state->registers[ecx] != 0 )
	    {
	      * (char*) state->registers[edi] = (b1 = * (char*) state->registers[esi]);

	      if ( state->efl & 0x00000400 )
		{
		  state->registers[edi]--;
		  state->registers[esi]--;
		}
	      else
		{
		  state->registers[edi]++;
		  state->registers[esi]++;
		}

	      if ( isprint(b1) )
		printf( "%*s  # -1[%%edi] = %02x  '%c'\n", n+IWIDTH, "", b1, b1 );
	      else
		printf( "%*s  # -1[%%edi] = %02x\n", n+IWIDTH, "", b1 );

	      state->registers[ecx]--;
	    }

	  printf( "%*s ## %%esi = %08x\n", n+IWIDTH, "", state->registers[esi] );
	  printf( "%*s    %%edi = %08x\n", n+IWIDTH, "", state->registers[edi] );
	}
      else
	{
	  * (char*) state->registers[edi] = (b1 = * (char*) state->registers[esi]);

	  if ( state->efl & 0x00000400 )
	    {
	      state->registers[edi]--;
	      state->registers[esi]--;
	    }
	  else
	    {
	      state->registers[edi]++;
	      state->registers[esi]++;
	    }

	  if ( isprint(b1) )
	    printf( "%*s ## -1[%%edi] = %02x  '%c'\n", n+IWIDTH, "", b1, b1 );
	  else
	    printf( "%*s ## -1[%%edi] = %02x\n", n+IWIDTH, "", b1 );
	  printf( "%*s    %%esi = %08x\n", n+IWIDTH, "", state->registers[esi] );
	  printf( "%*s    %%edi = %08x\n", n+IWIDTH, "", state->registers[edi] );
	}

      state->eip += 1;
      break;
  case movsx:
      r = ((struct modrm*)p)->reg;

      p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

      sprintf( ops, "%%%s,%s", regname32[r], op2 );

      printf( "%-*s    %%%s = %08x\n", IWIDTH, ops, regname32[r], state->registers[r] );
      printf( "%*s    %s = %0*x\n", n+IWIDTH, "", op2, size, srcv );

      srcv = fetchs( flags, src );
      printf( "%*s ## %%%s = %08x\n", n+IWIDTH, "", regname32[r], state->registers[r] = srcv );
      state->eip = (unsigned) p;
      break;
  case movzx:
      r = ((struct modrm*)p)->reg;

      p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

      sprintf( ops, "%%%s,%s", regname32[r], op2 );

      printf( "%-*s    %%%s = %08x\n", IWIDTH, ops, regname32[r], state->registers[r] );
      printf( "%*s    %s = %0*x\n", n+IWIDTH, "", op2, size, srcv );

      printf( "%*s ## %%%s = %08x\n", n+IWIDTH, "", regname32[r], state->registers[r] = srcv );
      state->eip = (unsigned) p;
      break;
  case neg:
      p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

      printf( "%-*s    %s = %0*x\n", IWIDTH, op1, op1, size, destv );

      destv = fetchs( flags, dest );
      asm( "movl %2,%%eax; negl %%eax; movl %%eax,%0; pushf; pop %1"
	   : "=g" (destv), "=m" (state->efl)
	   : "g" (destv)
	   : "%eax", "cc"
	 );
      store( flags, dest, destv );

      printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
      printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
      state->eip = (unsigned) p;
      break;
  case not:
      p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

      printf( "%-*s    %s = %0*x\n", IWIDTH, op1, op1, size, destv );

      destv = fetchs( flags, dest );
      asm( "movl %2,%%eax; notl %%eax; movl %%eax,%0; pushf; pop %1"
	   : "=g" (destv), "=m" (state->efl)
	   : "g" (destv)
	   : "%eax", "cc"
	 );
      store( flags, dest, destv );

      printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
      printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
      state->eip = (unsigned) p;
      break;
  case or:
      if ( flags & IMM8 )
        /* 83 */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  x = *(signed char*)p;
	  sprintf( ops, "%s,%d", op1, x );
	  printf( "%-*s    %s = %08x\n", IWIDTH, ops, op1, destv );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; orl %3,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv), "g" (x)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %08x\n", n+IWIDTH, "", op1, destv );
	  printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  state->eip = (unsigned) p + 1;
	}
      else
	if ( flags & IMM )
	  /* 0[cd], 8[01] */
	  {
	    if ( flags & MODRM )
	      /* 8[01] */
	      {
		p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

		x = fetchs( flags, p );
		sprintf( ops, "%s,%d", op1, x );
		printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

		destv = fetchs( flags, dest );
		asm( "movl %2,%%eax; orl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, dest, destv );

		printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip = (unsigned) p + (size >> 1);
	      }
	    else
	      /* 0[cd] */
	      {
		x = fetchs( flags, p );
		sprintf( ops, "%%%s,%d", reg, x );
		printf( "%-*s    %%eax = %08x\n", IWIDTH, ops, state->registers[eax] );

		destv = fetchs( flags, &state->registers[eax] );
		asm( "movl %2,%%eax; orl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, &state->registers[eax], destv );

		printf( "%*s ## %%eax = %08x\n", n+IWIDTH, "", state->registers[eax] );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip += 1 + (size >> 1);
	      }
	  }
	else
	  /* 0[89ab] */
	  {
	    p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	    sprintf( ops, "%s,%s", op1, op2 );
	    printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );
	    printf( "%*s    %s = %0*x\n", n+IWIDTH, "", op2, size, srcv );

	    destv = fetchs( flags, dest );
	    srcv = fetchs( flags, src );
	    asm( "movl %2,%%eax; orl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		 : "=g" (destv), "=m" (state->efl)
		 : "0" (destv), "g" (srcv)
		 : "%eax", "cc"
	       );
	    store( flags, dest, destv );

	    printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	    printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	    state->eip = (unsigned) p;
	  }

      break;
  case pop:
      if ( flags & OPD )
	{
	  printf( "%%%-*s    %%esp = %8p  ->  %08x\n", IWIDTH-1, reg,
	  					(unsigned*) state->registers[esp],
						((unsigned*) state->registers[esp])[0] );
	  state->registers[r] = * (unsigned*) state->registers[esp];
	  printf( "%*s ## %%%s = %8x\n", n+IWIDTH, "", reg, state->registers[r] );
	  state->eip += 1;
	}
      else
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );
	  printf( "%-*s    %%esp = %8p  ->  %08x\n", IWIDTH, op1, (unsigned*) state->registers[esp],
	  					((unsigned*) state->registers[esp])[0] );
	  *dest = * (unsigned*) state->registers[esp];
	  printf( "%*s ## %s = %8x\n", n+IWIDTH, "", op1, *dest );
	  state->eip = (unsigned) p;
	}
      state->registers[esp] += 4;
      break;
  case push:
      if ( flags & IMM )
        {
	  if ( flags & BYTE )
	    {
	      x = *(signed char*)p;
	      state->eip += 2;
	    }
	  else
	    {
	      x = *(int*)p;
	      state->eip += 5;
	    }
	  printf( "%-*d    %%esp = %8p\n", IWIDTH, x, (unsigned*) state->registers[esp] );
	}
      else
	if ( flags & OPD )
	  {
	    printf( "%%%-*s    %%esp = %8p\n", IWIDTH-1, reg, (unsigned*) state->registers[esp] );
	    x = state->registers[r];
	    state->eip += 1;
	  }
	else
	  {
	    p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );
	    x = destv;
	    printf( "%-*s    %%esp = %8p\n", IWIDTH, op1, (unsigned*) state->registers[esp] );
	    state->eip = (unsigned) p;
	  }

      * (unsigned*) (state->registers[esp] -= 4) = x;
      printf( "%*s ## %%esp = %8p  ->  %08x\n", n+IWIDTH, "",
      					(unsigned*) state->registers[esp],
					((unsigned*) state->registers[esp])[0] );
      break;
  case rep:
      printf( "\n" );
      prefix |= REP;
      state->eip += 1;
      return 0;
  case repne:
      printf( "\n" );
      prefix |= REPNE;
      state->eip += 1;
      return 0;
  case ret:
      if ( flags & IMM16 )
        {
	  x = *(unsigned short*)p;
	  printf( "%-*d    %%esp = %8p  ->  %08x %08x\n", IWIDTH, x,
	  					(unsigned*) state->registers[esp],
						((unsigned*) state->registers[esp])[0],
						((unsigned*) state->registers[esp])[1] );
	}
      else
        {
	  x = 0;
	  printf( "%*s    %%esp = %8p  ->  %08x %08x\n", IWIDTH, "",
	  					(unsigned*) state->registers[esp],
						((unsigned*) state->registers[esp])[0],
						((unsigned*) state->registers[esp])[1] );
	}
      state->eip = * (unsigned*) state->registers[esp];
      state->registers[esp] += x + sizeof(int);
      printf( "%*s ## %%esp = %8p\n", n+IWIDTH, "", (unsigned*) state->registers[esp] );
      printf( "%*s    eip = %8p\n", n+IWIDTH, "", (unsigned*) state->eip );
      break;
  case sar:
      p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

      if ( ! (flags & (MODRM | IMM)) )
	/* d[01] */
	{
	  sprintf( ops, "%s,1", op1 );
	  printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; sarl $1,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	  printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  state->eip = (unsigned) p;
	}
      else
	{
	  if ( flags & IMM )
	    /* c[01] */
	    {
	      b1 = *(unsigned char*)p;
	      sprintf( ops, "%s,%u", op1, b1 );
	      state->eip = (unsigned) p + 1;
	    }
	  else
	    /* d[23] */
	    {
	      b1 = state->registers[ecx];
	      sprintf( ops, "%s,%%cl", op1 );
	      state->eip = (unsigned) p;
	    }

	  printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );
	  if ( ! (flags & IMM) )
	    printf( "%*s    %%cl = %u\n", n+IWIDTH, "", b1 );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; movb %3,%%cl; sarl %%cl,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv), "g" (b1)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	  printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	}
      break;
  case sbb:
      if ( flags & IMM8 )
        /* 83 */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  x = *(signed char*)p;
	  sprintf( ops, "%s,%d", op1, x );
	  printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; sbbl %3,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv), "g" (x)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	  printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  state->eip = (unsigned) p + 1;
	}
      else
	if ( flags & IMM )
	  /* 1[cd], 8[01] */
	  {
	    if ( flags & MODRM )
	      /* 8[01] */
	      {
		p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

		x = fetchs( flags, p );
		sprintf( ops, "%s,%d", op1, x );
		printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

		destv = fetchs( flags, dest );
		asm( "movl %2,%%eax; sbbl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, dest, destv );

		printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip = (unsigned) p + (size >> 1);
	      }
	    else
	      /* 1[cd] */
	      {
		x = fetchs( flags, p );
		sprintf( ops, "%%%s,%d", reg, x );
		printf( "%-*s    %%eax = %08x\n", IWIDTH, ops, state->registers[eax] );

		destv = fetchs( flags, &state->registers[eax] );
		asm( "movl %2,%%eax; sbbl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, &state->registers[eax], destv );

		printf( "%*s ## %%eax = %08x\n", n+IWIDTH, "", state->registers[eax] );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip += 1 + (size >> 1);
	      }
	  }
	else
	  /* 1[89ab] */
	  {
	    p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	    sprintf( ops, "%s,%s", op1, op2 );
	    printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );
	    printf( "%*s    %s = %0*x\n", n+IWIDTH, "", op2, size, srcv );

	    destv = fetchs( flags, dest );
	    srcv = fetchs( flags, src );
	    asm( "movl %2,%%eax; sbbl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		 : "=g" (destv), "=m" (state->efl)
		 : "0" (destv), "g" (srcv)
		 : "%eax", "cc"
	       );
	    store( flags, dest, destv );

	    printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	    printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	    state->eip = (unsigned) p;
	  }

      break;
  case scasb:
      if ( prefix & (REP | REPNE) )
        {
	  if ( state->registers[ecx] != 0 )
	    {
	      b2 = state->registers[eax];

	      printf( "%*s    %%al = %02x\n", IWIDTH, "", b2 );

	      while( state->registers[ecx] != 0 )
	        {
		  printf( "%*s    %%ecx = %08x\n", n+IWIDTH, "", state->registers[ecx] );
		  printf( "%*s    [%%edi] = %02x\n", n+IWIDTH, "",
		  					* (unsigned char*) state->registers[edi] );

		  b1 = fetch( flags, state->registers[edi] );
		  asm( "movb %1,%%al; cmpb %%al,%2; pushf; pop %0"
		       : "=m" (state->efl)
		       : "g" (b1), "g" (b2)
		       : "%eax", "cc"
		     );

		  printf( "%*s  # efl: %s\n", n+IWIDTH, "", theflags(state->efl) );

		  if ( state->efl & 0x00000400 )
		    state->registers[edi]--;
		  else
		    state->registers[edi]++;

		  state->registers[ecx]--;

		  if ( prefix & REP && ! (state->efl & 0x00000040) )
		    break;
		  if ( prefix & REPNE && state->efl & 0x00000040 )
		    break;
		}

	      printf( "%*s ## %%edi = %08x\n", n+IWIDTH, "", state->registers[edi] );
	    }
	  else
	    printf( "%*s    %%ecx = 0\n", IWIDTH, "" );
	}
      else
	{
	  printf( "%*s    %%al = %02x\n", IWIDTH, "", state->registers[eax] & 0xff );
	  printf( "%*s    [%%edi] = %02x\n", IWIDTH, "", * (unsigned char*) state->registers[edi] );

	  b1 = fetch( flags, state->registers[edi] );
	  b2 = state->registers[eax];
	  asm( "movb %1,%%al; cmpb %%al,%2; pushf; pop %0"
	       : "=m" (state->efl)
	       : "g" (b1), "g" (b2)
	       : "%eax", "cc"
	     );

	  if ( state->efl & 0x00000400 )
	    state->registers[edi]--;
	  else
	    state->registers[edi]++;

	  printf( "%*s ## efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  printf( "%*s    %%edi = %08x\n", n+IWIDTH, "", state->registers[edi] );
	}

      state->eip += 1;
      break;
  case setcc:
      r = p[-1] & 0x0f;

      p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

      printf( "%-*s    efl: %s\n", IWIDTH, op1, theflags(state->efl) );

      switch( r )
        {
      case 0:
	  asm( "push %1; popf; seto %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 1:
	  asm( "push %1; popf; setno %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 2:
	  asm( "push %1; popf; setb %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 3:
	  asm( "push %1; popf; setae %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 4:
	  asm( "push %1; popf; setz %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 5:
	  asm( "push %1; popf; setnz %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 6:
	  asm( "push %1; popf; setbe %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 7:
	  asm( "push %1; popf; seta %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 8:
	  asm( "push %1; popf; sets %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 9:
	  asm( "push %1; popf; setns %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 10:
	  asm( "push %1; popf; setpe %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 11:
	  asm( "push %1; popf; setpo %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 12:
	  asm( "push %1; popf; setl %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 13:
	  asm( "push %1; popf; setge %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 14:
	  asm( "push %1; popf; setle %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
      case 15:
	  asm( "push %1; popf; setg %%al; movb %%al,%0"
	       : "=g" (b1)
	       : "m" (state->efl)
	       : "%eax", "cc"
	     );
          break;
	}
      printf( "%*s ## %s = %02x\n", n+IWIDTH, "", op1, * (char*) dest = b1 );
      state->eip = (unsigned) p;
      break;
  case stc:
      state->efl |= 0x00000001;
      printf( "%*s ## efl: %s\n", IWIDTH, "", theflags(state->efl) );
      state->eip += 1;
      break;
  case std:
      state->efl |= 0x00000400;
      printf( "%*s ## efl: %s\n", IWIDTH, "", theflags(state->efl) );
      state->eip += 1;
      break;
  case sub:
      if ( flags & IMM8 )
        /* 83 */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  x = *(signed char*)p;
	  sprintf( ops, "%s,%d", op1, x );
	  printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; subl %3,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv), "g" (x)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	  printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  state->eip = (unsigned) p + 1;
	}
      else
	if ( flags & IMM )
	  /* 2[cd], 8[01] */
	  {
	    if ( flags & MODRM )
	      /* 8[01] */
	      {
		p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

		x = fetchs( flags, p );
		sprintf( ops, "%s,%d", op1, x );
		printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

		destv = fetchs( flags, dest );
		asm( "movl %2,%%eax; subl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, dest, destv );

		printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip = (unsigned) p + (size >> 1);
	      }
	    else
	      /* 2[cd] */
	      {
		x = fetchs( flags, p );
		sprintf( ops, "%%%s,%d", reg, x );
		printf( "%-*s    %%eax = %08x\n", IWIDTH, ops, state->registers[eax] );

		destv = fetchs( flags, &state->registers[eax] );
		asm( "movl %2,%%eax; subl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, &state->registers[eax], destv );

		printf( "%*s ## %%eax = %08x\n", n+IWIDTH, "", state->registers[eax] );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip += 1 + (size >> 1);
	      }
	  }
	else
	  /* 2[89ab] */
	  {
	    p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	    sprintf( ops, "%s,%s", op1, op2 );
	    printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );
	    printf( "%*s    %s = %0*x\n", n+IWIDTH, "", op2, size, srcv );

	    destv = fetchs( flags, dest );
	    srcv = fetchs( flags, src );
	    asm( "movl %2,%%eax; subl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		 : "=g" (destv), "=m" (state->efl)
		 : "0" (destv), "g" (srcv)
		 : "%eax", "cc"
	       );
	    store( flags, dest, destv );

	    printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	    printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	    state->eip = (unsigned) p;
	  }

      break;
  case shl:
      p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

      if ( ! (flags & (MODRM | IMM)) )
	/* d[01] */
	{
	  sprintf( ops, "%s,1", op1 );

	  printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; shll $1,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	  printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  state->eip = (unsigned) p;
	}
      else
	{
	  if ( flags & IMM )
	    /* c[01] */
	    {
	      b1 = *(unsigned char*)p;
	      sprintf( ops, "%s,%u", op1, b1 );
	      state->eip = (unsigned) p + 1;
	    }
	  else
	    /* d[23] */
	    {
	      b1 = state->registers[ecx];
	      sprintf( ops, "%s,%%cl", op1 );
	      state->eip = (unsigned) p;
	    }

	  printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );
	  if ( ! (flags & IMM) )
	    printf( "%*s    %%cl = %u\n", n+IWIDTH, "", b1 );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; movb %3,%%cl; shll %%cl,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv), "g" (b1)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	  printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	}
      break;
  case shr:
      p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

      if ( ! (flags & (MODRM | IMM)) )
	/* d[01] */
	{
	  sprintf( ops, "%s,1", op1 );

	  printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; shrl $1,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	  printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  state->eip = (unsigned) p;
	}
      else
	{
	  if ( flags & IMM )
	    /* c[01] */
	    {
	      b1 = *(unsigned char*)p;
	      sprintf( ops, "%s,%u", op1, b1 );
	      state->eip = (unsigned) p + 1;
	    }
	  else
	    /* d[23] */
	    {
	      b1 = state->registers[ecx];
	      sprintf( ops, "%s,%%cl", op1 );
	      state->eip = (unsigned) p;
	    }

	  printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );
	  if ( ! (flags & IMM) )
	    printf( "%*s    %%cl = %u\n", n+IWIDTH, "", b1 );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; movb %3,%%cl; shrl %%cl,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv), "g" (b1)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	  printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	}
      break;
  case test:
      if ( flags & IMM )
	/* a[89], f[67] */
	{
	  if ( flags & MODRM )
	    /* f[67] */
	    {
	      p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	      x = fetchs( flags, p );
	      sprintf( ops, "%s,%d", op1, x );
	      printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

	      destv = fetchs( flags, dest );
	      asm( "movl %1,%%eax; testl %2,%%eax; pushf; pop %0"
		   : "=m" (state->efl)
		   : "g" (destv), "g" (x)
		   : "%eax", "cc"
		 );

	      printf( "%*s      ## efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	      state->eip = (unsigned) p + (size >> 1);
	    }
	  else
	    /* a[89] */
	    {
	      x = fetchs( flags, p );
	      sprintf( ops, "%%%s,%d", reg, x );
	      printf( "%-*s    %%eax = %08x\n", IWIDTH, ops, state->registers[eax] );

	      destv = fetchs( flags, &state->registers[eax] );
	      asm( "movl %1,%%eax; testl %2,%%eax; pushf; pop %0"
		   : "=m" (state->efl)
		   : "g" (destv), "g" (x)
		   : "%eax", "cc"
		 );

	      printf( "%*s      ## efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	      state->eip += 1 + (size >> 1);
	    }
	}
      else
	/* 8[45] */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	  sprintf( ops, "%s,%s", op1, op2 );
	  printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );
	  printf( "%*s    %s = %0*x\n", n+IWIDTH, "", op2, size, srcv );

	  destv = fetchs( flags, dest );
	  srcv = fetchs( flags, src );
	  asm( "movl %1,%%eax; testl %2,%%eax; pushf; pop %0"
	       : "=m" (state->efl)
	       : "g" (destv), "g" (srcv)
	       : "%eax", "cc"
	     );

	  printf( "%*s      ## efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  state->eip = (unsigned) p;
	}

      break;
  case xchg:
      if ( flags & OPD )
        {
	  printf( "%%%s,%%%-*s    %%%s = %08x\n", "eax", IWIDTH-6, reg, reg, state->registers[r] );
	  printf( "%*s    %%eax = %08x\n", n+IWIDTH, "", state->registers[eax] );
	  x = state->registers[r];
	  state->registers[r] = state->registers[eax];
	  state->registers[eax] = x;
	  state->eip += 1;
	  printf( "%*s    %%%s = %08x\n", n+IWIDTH, "", reg, state->registers[r] );
	  printf( "%*s    %%eax = %08x\n", n+IWIDTH, "", state->registers[eax] );
	}
      else
        {
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	  sprintf( ops, "%s,%s", op1, op2 );
	  printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );
	  printf( "%*s    %s = %0*x\n", n+IWIDTH, "", op2, size, srcv );

	  store( flags, dest, srcv );
	  store( flags, src, destv );

	  printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, srcv );
	  printf( "%*s    %s = %0*x\n", n+IWIDTH, "", op2, size, destv );
	  state->eip = (unsigned) p;
	}
      break;
  case xor:
      if ( flags & IMM8 )
        /* 83 */
	{
	  p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

	  x = *(signed char*)p;
	  sprintf( ops, "%s,%d", op1, x );
	  printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

	  destv = fetchs( flags, dest );
	  asm( "movl %2,%%eax; xorl %3,%%eax; movl %%eax,%0; pushf; pop %1"
	       : "=g" (destv), "=m" (state->efl)
	       : "0" (destv), "g" (x)
	       : "%eax", "cc"
	     );
	  store( flags, dest, destv );

	  printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	  printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	  state->eip = (unsigned) p + 1;
	}
      else
	if ( flags & IMM )
	  /* 3[45], 8[01] */
	  {
	    if ( flags & MODRM )
	      /* 8[01] */
	      {
		p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, NULL, NULL, NULL, state );

		x = fetchs( flags, p );
		sprintf( ops, "%s,%d", op1, x );
		printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );

		destv = fetchs( flags, dest );
		asm( "movl %2,%%eax; xorl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, dest, destv );

		printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip = (unsigned) p + (size >> 1);
	      }
	    else
	      /* 3[45] */
	      {
		x = fetchs( flags, p );
		sprintf( ops, "%%%s,%d", reg, x );
		printf( "%-*s    %%eax = %08x\n", IWIDTH, ops, state->registers[eax] );

		destv = fetchs( flags, &state->registers[eax] );
		asm( "movl %2,%%eax; xorl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		     : "=g" (destv), "=m" (state->efl)
		     : "0" (destv), "g" (x)
		     : "%eax", "cc"
		   );
		store( flags, &state->registers[eax], destv );

		printf( "%*s ## %%eax = %08x\n", n+IWIDTH, "", state->registers[eax] );
		printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
		state->eip += 1 + (size >> 1);
	      }
	  }
	else
	  /* 3[0123] */
	  {
	    p = modrm( flags, (struct modrm*)p, &dest, &destv, op1, &src, &srcv, op2, state );

	    sprintf( ops, "%s,%s", op1, op2 );
	    printf( "%-*s    %s = %0*x\n", IWIDTH, ops, op1, size, destv );
	    printf( "%*s    %s = %0*x\n", n+IWIDTH, "", op2, size, srcv );

	    destv = fetchs( flags, dest );
	    srcv = fetchs( flags, src );
	    asm( "movl %2,%%eax; xorl %3,%%eax; movl %%eax,%0; pushf; pop %1"
		 : "=g" (destv), "=m" (state->efl)
		 : "0" (destv), "g" (srcv)
		 : "%eax", "cc"
	       );
	    store( flags, dest, destv );

	    printf( "%*s ## %s = %0*x\n", n+IWIDTH, "", op1, size, destv );
	    printf( "%*s    efl: %s\n", n+IWIDTH, "", theflags(state->efl) );
	    state->eip = (unsigned) p;
	  }

      break;
  default:
      fault( state, SIGIOT );
      return -1;
    }

  prefix = 0;
  return 0;
}


/*
 * Compute effective address from the mod r/m byte
 */
 unsigned char*
modrm( unsigned flags, struct modrm *rmp, unsigned **dest, unsigned *destv, char *op1,
					  unsigned **src, unsigned *srcv, char *op2,
					  struct iAPX486 *state )
{
  unsigned t, *tt;
  unsigned char *rc;
  char o[256];

  if ( src != NULL )
    {
      if ( flags & BYTE )
        {
	  *src = (unsigned*) ((char*) &state->registers[rmp->reg & 0x03] +
	  						((rmp->reg & 0x04)? 1 : 0));
	  sprintf( op2, "%%%s", regname8[rmp->reg] );
	}
      else
        {
	  *src = &state->registers[rmp->reg];

	  if ( flags & WORD )
	    sprintf( op2, "%%%s", regname16[rmp->reg] );
	  else
	    sprintf( op2, "%%%s", regname32[rmp->reg] );
	}

      if ( ! (flags & NOV) )
        *srcv = fetch( flags, *src );
    }

  switch( rmp->mod )
    {
  case 0:
      if ( rmp->rm == esp )
	rc = sib( flags, 0, (struct sib *)((char*) rmp + 1), dest, destv, op1, state );
      else
	{
	  if ( rmp->rm == ebp )
	    {
	      *dest = * (unsigned**) ((char*) rmp + 1);
	      sprintf( op1, "%p", *dest );

	      rc = (unsigned char*) rmp + 5;
	    }
	  else
	    {
	      *dest = (unsigned*) state->registers[rmp->rm];
	      sprintf( op1, "[%%%s]", regname32[rmp->rm] );
	      
	      rc = (unsigned char*) rmp + 1;
	    }
	}

      if ( ! (flags & NOV) )
        *destv = fetch( flags, *dest );

      break;
  case 1:
      if ( rmp->rm == esp )
	rc = sib( flags, 1, (struct sib *)((char*) rmp + 1), dest, destv, op1, state );
      else
	{
	  *dest = (unsigned*) (state->registers[rmp->rm] + ((signed char*) rmp)[1]);
	  sprintf( op1, "%d[%%%s]", * (signed char*) ((char*) rmp + 1), regname32[rmp->rm] );
	  
	  rc = (unsigned char*) rmp + 2;
	}

      if ( ! (flags & NOV) )
        *destv = fetch( flags, *dest );

      break;
  case 2:
      if ( rmp->rm == esp )
	rc = sib( flags, 2, (struct sib *)((char*) rmp + 1), dest, destv, op1, state );
      else
	{
	  *dest = (unsigned*) (state->registers[rmp->rm] + * (int*) ((char*) rmp + 1));
	  sprintf( op1, "%d[%%%s]", * (int*) ((char*) rmp + 1), regname32[rmp->rm] );
	  
	  rc = (unsigned char*) rmp + 5;
	}

      if ( ! (flags & NOV) )
        *destv = fetch( flags, *dest );

      break;
  case 3:
      if ( flags & BYTE )
	{
	  *dest = (unsigned*) ((char*) &state->registers[rmp->rm & 0x03] +
	  						((rmp->rm & 0x04)? 1 : 0));
	  sprintf( op1, "%%%s", regname8[rmp->rm] );
	}
      else
	{
	  *dest = &state->registers[rmp->rm];
	  if ( flags & WORD )
	    sprintf( op1, "%%%s", regname16[rmp->rm] );
	  else
	    sprintf( op1, "%%%s", regname32[rmp->rm] );
	}

      if ( ! (flags & NOV) )
        *destv = fetch( flags, *dest );
      
      rc = (unsigned char*) rmp + 1;
      break;
    }

  if ( flags & MR )
    {
      if ( src != NULL )
        {
	  tt = *src;
	  *src = *dest;
	  *dest = tt;

	  if ( ! (flags & NOV) )
	    {
	      t = *srcv;
	      *srcv = *destv;
	      *destv = t;
	    }

	  strcpy( o, op1 );
	  strcpy( op1, op2 );
	  strcpy( op2, o );
	}
    }

  return rc;
}


/*
 * Compute effective address from the sib byte
 */
 unsigned char*
sib( unsigned flags, unsigned mod, struct sib *sibp, unsigned **dest, unsigned *destv, char *op,
			struct iAPX486 *state )
{
  unsigned x;
  unsigned char *rc;

  if ( sibp->index == esp )
    {
      fault( state, SIGIOT );
      return NULL;
    }

  x = 1 << sibp->ss;

  if ( sibp->base == ebp && mod == 0 )
    {
      *dest = (unsigned*) (x * state->registers[sibp->index] + * (int*) ((char*) sibp + 1));
      sprintf( op, "%d[%%%s*%d,-]", * (int*) ((char*) sibp + 1), regname32[sibp->index], x );
      
      rc = (unsigned char*) sibp + 5;
    }
  else
    switch( mod )
      {
    case 0:
	*dest = (unsigned*) (state->registers[sibp->base] + x * state->registers[sibp->index]);
	sprintf( op, "[%%%s*%d,%%%s]", regname32[sibp->index], x, regname32[sibp->base] );
	
	rc = (unsigned char*) sibp + 1;
	break;
    case 1:
	*dest = (unsigned*) (state->registers[sibp->base] + x * state->registers[sibp->index] +
								((signed char*) sibp)[1]);
	sprintf( op, "%d[%%%s*%d,%%%s]", ((signed char*)sibp)[1], regname32[sibp->index], x,
								regname32[sibp->base] );
	
	rc = (unsigned char*) sibp + 2;
	break;
    case 2:
	*dest = (unsigned*) (state->registers[sibp->base] + x * state->registers[sibp->index] +
								* (int*) ((char*) sibp + 1));
	sprintf( op, "%d[%%%s*%d,%%%s]", * (int*) ((char*) sibp + 1), regname32[sibp->index], x,
								regname32[sibp->base] );
	
	rc = (unsigned char*) sibp + 5;
	break;
      }

  return rc;
}


/*
 * Format the flags for printing
 */
 char*
theflags( unsigned flags )
{
  static char buffer[100];

  sprintf( buffer, "of=%d df=%d sf=%d zf=%d af=%d pf=%d cf=%d",
			(flags & 0x00000800)? 1 : 0,
			(flags & 0x00000400)? 1 : 0,
			(flags & 0x00000080)? 1 : 0,
			(flags & 0x00000040)? 1 : 0,
			(flags & 0x00000010)? 1 : 0,
			(flags & 0x00000004)? 1 : 0,
			(flags & 0x00000001)? 1 : 0 );

  return buffer;
}


/*
 * Processor fault
 */
 void
fault( struct iAPX486 *state, int sig )
{
  struct sigframe sigframe;

  sigframe.sf_signum = sig;
  sigframe.sf_code = 0;
  sigframe.sf_scp = &sigframe.sf_sc;
  sigframe.sf_addr = NULL;
  sigframe.sf_handler = NULL;
  sigframe.sf_sc.sc_onstack = 0;
  sigframe.sf_sc.sc_mask = 0;
  sigframe.sf_sc.sc_esp = state->registers[esp];
  sigframe.sf_sc.sc_ebp = state->registers[ebp];
  sigframe.sf_sc.sc_isp = 0;
  sigframe.sf_sc.sc_eip = state->eip;
  sigframe.sf_sc.sc_efl = state->efl;
  sigframe.sf_sc.sc_es = 0;
  sigframe.sf_sc.sc_ds = 0;
  sigframe.sf_sc.sc_cs = 0;
  sigframe.sf_sc.sc_ss = 0;
  sigframe.sf_sc.sc_edi = state->registers[edi];
  sigframe.sf_sc.sc_esi = state->registers[esi];
  sigframe.sf_sc.sc_ebx = state->registers[ebx];
  sigframe.sf_sc.sc_edx = state->registers[edx];
  sigframe.sf_sc.sc_ecx = state->registers[ecx];
  sigframe.sf_sc.sc_eax = state->registers[eax];

  abend( 0, &sigframe );
}

/* vi: set sw=2 ai sm: */
