/*
 * This program reformats program file generated by the XCOM compiler to allow
 * program loading from different direct access devices.  One advatntage is the
 * possibility of using extremely fast devices such as 2301 drums for compiler
 * residence while using small slow disks like 2311's for scratch during the
 * compiler generation process.
 *
 * The principle effect of this program is to copy a standard format file from
 * the dataset 'sysut1' to the dataset 'sysut2'.  The DD cards for both data
 * sets must specify record length and record format
 *
 * A side effect of this program is to rebuild the file control information
 * which is at the head of the program file.
 *
 *
 * David B. Wortman
 * Stanford University
 * March  1969
 *
 *
 * Translation to C
 *
 * Copyright 1996, 1997, David E. Bodenstab
 * All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

 inline
 int
max( int x, int y )
{
  if ( x > y )
    return x;
  else
    return y;
}


/*
 * DSECT which defines the format of binary program control information and the
 * starting point for programs
 */
struct FILECTRL
  {
    int BYTSCODE;   /* NUMBER OF BYTES OF CODE */
    int BYTSDATA;   /* NUMBER OF BYTES OF DATA AREA */
    int BLKSCODE;   /* NUMBER OF RECORDS OF CODE */
    int BLKSDATA;   /* NUMBER OF RECORDS OF DATA AREA */
    int BYTSBLK;    /* BLOCKSIZE OF THE XPL PROGRAM FILE */
    int BYTSFULL;   /* NUMBER OF BYTES OF CODE ACTUALLY USED IN LAST CODE RECORD */
    int DATABYTS;   /* NUMBER OF BYTES OF DATA ACTUALLY USED IN LAST DATA RECORD */
    int pad[8];     /* REMAINDER OF THE CONTROL BLOCK IS UNUSED */
  };


struct word
  {
    unsigned char w[4];
  };



void	storew( struct word*, int );
int	fetchw( struct word* );
void	pctrl( char*, struct FILECTRL* );

 int
main( int argc, char **argv )
{
  FILE *sysut1, *sysut2;
  struct FILECTRL filectrl, *fp;
  char *buffer, *r2;
  int c, count, rc;
  int LRECL = -1, OLDDBL, OLDBBL, NEWDBL;
  div_t qr;

  while( (c = getopt(argc,argv,"l:")) != EOF )
    switch( c )
      {
    default:
	fprintf( stderr, "Usage: %s [-l LRECL] infile outfile\n", argv[0] );
	fprintf( stderr, "       %s infile\n", argv[0] );
	exit( 1 );
    case 'l':
	LRECL = atol( optarg );
	break;
      }
  if ( ! ((LRECL == -1 && argc == 2) || (argc - optind == 2)) )
    {
      fprintf( stderr, "Usage: %s [-l LRECL] infile outfile\n", argv[0] );
      fprintf( stderr, "       %s infile\n", argv[0] );
      exit( 1 );
    }

  if ( (sysut1 = fopen(argv[optind],"r")) == NULL )
    {
      fprintf( stderr, "%s: %s: cannot open (%s)\n", argv[0],
						    argv[optind], strerror(errno) );
      exit( 1 );
    }

  if ( (rc = fread(&filectrl,1,sizeof(filectrl),sysut1)) != sizeof(filectrl) )
    {
      fprintf( stderr, "%s: fread(\"%s\",%d) returned %d\n", argv[0],
      						argv[optind], sizeof(filectrl), rc );
      exit( 1 );
    }

  rewind( sysut1 );

  pctrl( "OLD", &filectrl );

  if ( argc - optind == 2 )
    {
      if ( LRECL == -1 )
        LRECL = 512;

      OLDDBL = fetchw( (struct word*)&filectrl.BLKSDATA );
      OLDBBL = fetchw( (struct word*)&filectrl.BYTSBLK );

      count = max( fetchw((struct word*)&filectrl.BYTSCODE),
                   fetchw((struct word*)&filectrl.BYTSDATA) );
      if ( (buffer = malloc(count)) == NULL )
	{
	  fprintf( stderr, "%s: malloc(%u) failed\n", argv[0], count );
	  exit( 1 );
	}

      for( r2 = buffer, count = fetchw((struct word*)&filectrl.BLKSCODE);
           count > 0;
	   --count, r2 += OLDBBL )
	{
	  if ( (rc = fread(r2,1,OLDBBL,sysut1)) != OLDBBL )
	    {
	      fprintf( stderr, "%s: fread(\"%s\",%d) returned %d\n", argv[0],
	      						argv[optind], OLDBBL, rc );
	      exit( 1 );
	    }
	}

      fp = (struct FILECTRL*) buffer;

      /*
       * Compute the new file control block
       */
      storew( (struct word*)&fp->BYTSBLK, LRECL );

      qr = div( fetchw( (struct word*)&fp->BYTSCODE ) - OLDBBL +
      					fetchw( (struct word*)&fp->BYTSFULL ), LRECL );

      if ( qr.rem == 0 )
	storew( (struct word*)&fp->BYTSFULL, LRECL );
      else
	{
	  storew( (struct word*)&fp->BYTSFULL, qr.rem );
	  qr.quot += 1;
	}
      storew( (struct word*)&fp->BLKSCODE, qr.quot );

      storew( (struct word*)&fp->BYTSCODE, qr.quot * LRECL );

      qr = div( fetchw( (struct word*)&fp->BYTSDATA ) - OLDBBL +
      					fetchw( (struct word*)&fp->DATABYTS ), LRECL );

      if ( qr.rem == 0 )
	storew( (struct word*)&fp->DATABYTS, LRECL );
      else
	{
	  storew( (struct word*)&fp->DATABYTS, qr.rem );
	  qr.quot += 1;
	}
      storew( (struct word*)&fp->BLKSDATA, qr.quot );

      storew( (struct word*)&fp->BYTSDATA, (NEWDBL = qr.quot) * LRECL );

      pctrl( "NEW", fp );

      /*
       * Write out the code blocks in the new format
       */
      if ( (sysut2 = fopen(argv[optind+1],"w")) == NULL )
	{
	  fprintf( stderr, "%s: %s: cannot open (%s)\n", argv[0], argv[optind+1], strerror(errno) );
	  unlink( argv[optind+1] );
	  exit( 1 );
	}

      for( r2 = buffer, count = fetchw((struct word*)&fp->BLKSCODE);
           count > 0;
	   --count, r2 += LRECL )
	{
	  if ( (rc = fwrite(r2,1,LRECL,sysut2)) != LRECL )
	    {
	      fprintf( stderr, "%s: fwrite(\"%s\",%d) returned %d\n", argv[0],
	      					argv[optind+1], LRECL, rc );
	      unlink( argv[optind+1] );
	      exit( 1 );
	    }
	}

      /*
       * Read in the data blocks for the xpl binary program on sysut1
       */
      for( r2 = buffer, count = OLDDBL; count > 0; --count, r2 += OLDBBL )
	{
	  if ( (rc = fread(r2,1,OLDBBL,sysut1)) != OLDBBL )
	    {
	      fprintf( stderr, "%s: fread(\"%s\",%d) returned %d\n", argv[0],
	      					argv[optind], OLDBBL, rc );
	      unlink( argv[optind+1] );
	      exit( 1 );
	    }
	}

      if ( (c = fgetc(sysut1)) != EOF || ! feof(sysut1) )
        {
	  fprintf( stderr, "%s: file \"%s\" is not at EOF\n", argv[0], argv[optind] );
	  unlink( argv[optind+1] );
	  exit( 1 );
	}

      /*
       * Write out the data blocks in the new format
       */
      for( r2 = buffer, count = NEWDBL; count > 0; --count, r2 += LRECL )
	{
	  if ( (rc = fwrite(r2,1,LRECL,sysut2)) != LRECL )
	    {
	      fprintf( stderr, "%s: fwrite(\"%s\",%d) returned %d\n", argv[0],
	      					argv[optind+1], LRECL, rc );
	      unlink( argv[optind+1] );
	      exit( 1 );
	    }
	}

      /*
       * Free memory, close all files and return to OS/360
       */
      fclose( sysut2 );

      free( buffer );
    }

  fclose( sysut1 );

  return 0;
}



/*
 * Routine to print the file control blocks in hex
 */
 void
pctrl( char *name, struct FILECTRL *fp )
{
  printf( "%s FILE CONTROL BLOCK    %08.8X  %08.8X  %08.8X  %08.8X  %08.8X  %08.8X  %08.8X\n",
	  name,
	  fetchw((struct word*)&fp->BYTSCODE),
	  fetchw((struct word*)&fp->BYTSDATA),
	  fetchw((struct word*)&fp->BLKSCODE),
	  fetchw((struct word*)&fp->BLKSDATA),
	  fetchw((struct word*)&fp->BYTSBLK),
	  fetchw((struct word*)&fp->BYTSFULL),
	  fetchw((struct word*)&fp->DATABYTS) );
}



/*
 * Routines to convert between native and IBM/360 binary integers
 */
 void
storew( struct word *addr, int word )
{
  register char *t = (char*)&word;

  addr->w[0] = t[3];
  addr->w[1] = t[2];
  addr->w[2] = t[1];
  addr->w[3] = t[0];
}

 int
fetchw( struct word *addr )
{
  int word;
  register char *t = (char*)&word;

  t[3] = addr->w[0];
  t[2] = addr->w[1];
  t[1] = addr->w[2];
  t[0] = addr->w[3];
  return( word );
}
